# Copyright (C) 2017 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import String

from os_ken.lib import netdevice
from os_ken.lib import ip
from os_ken.lib.packet import zebra

from . import base


LOG = logging.getLogger(__name__)

# Default value for ethernet interface
DEFAULT_ETH_FLAGS = (
    netdevice.IFF_UP
    | netdevice.IFF_BROADCAST
    | netdevice.IFF_RUNNING
    | netdevice.IFF_MULTICAST)
DEFAULT_ETH_MTU = 1500


class Interface(base.Base):
    """
    Interface table for Zebra protocol service.

    The default value for each fields suppose "Loopback" interface.

    ``ifindex``: Number of index.

    ``ifname``: Name of this interface.

    ``status``: A combination of flags
    "os_ken.lib.packet.zebra.ZEBRA_INTERFACE_*".
    The default value shows "active" and "link-detect".

    ``flags``: A combination of flags "os_ken.lib.netdevice.IFF_*".
    The default value show "up", "loopback" and "running".

    ``metric``: Metric of this interface.

    ``ifmtu``: IPv4 MTU of this interface.

    ``ifmtu6``: IPv6 MTU of this interface.

    ``bandwidth``: Bandwidth of this interface.

    ``ll_type``: Link Layer Type.
    One of "os_ken.lib.packet.zebra.ZEBRA_LLT_*" types.

    ``hw_addr``: Hardware address of this interface (mostly, MAC address).

    ``inet``: List of IPv4 addresses separated by a comma.
    (e.g., "192.168.1.100/24,192.168.2.100/24)".

    ``inet6``: List of IPv6 addresses separated by a comma.
    """
    __tablename__ = 'interface'

    ifindex = Column(Integer, primary_key=True)
    ifname = Column(String, default="lo")
    status = Column(
        Integer,
        default=(
            zebra.ZEBRA_INTERFACE_ACTIVE
            | zebra.ZEBRA_INTERFACE_LINKDETECTION))
    flags = Column(
        Integer,
        default=(
            netdevice.IFF_UP
            | netdevice.IFF_LOOPBACK
            | netdevice.IFF_RUNNING))
    metric = Column(Integer, default=1)
    ifmtu = Column(Integer, default=0x10000)
    ifmtu6 = Column(Integer, default=0x10000)
    bandwidth = Column(Integer, default=0)
    ll_type = Column(Integer, default=zebra.ZEBRA_LLT_ETHER)
    hw_addr = Column(String, default='00:00:00:00:00:00')
    # Note: Only the PostgreSQL backend has support sqlalchemy.ARRAY,
    # we use the comma separated string as array instead.
    inet = Column(String, default='')
    inet6 = Column(String, default='')


@base.sql_function
def ip_link_show(session, **kwargs):
    """
    Returns a first interface record matching the given filtering rules.

    The arguments for "kwargs" is the same with Interface class.

    :param session: Session instance connecting to database.
    :param kwargs: Filtering rules to query.
    :return: An instance of Interface record.
    """
    return session.query(Interface).filter_by(**kwargs).first()


@base.sql_function
def ip_link_show_all(session, **kwargs):
    """
    Returns all interface records matching the given filtering rules.

    The arguments for "kwargs" is the same with Interface class.

    :param session: Session instance connecting to database.
    :param kwargs: Filtering rules to query.
    :return: A list of Interface records.
    """
    return session.query(Interface).filter_by(**kwargs).all()


@base.sql_function
def ip_link_add(session, name, type_='loopback', lladdr='00:00:00:00:00:00'):
    """
    Adds an interface record into Zebra protocol service database.

    The arguments are similar to "ip link add" command of iproute2.

    :param session: Session instance connecting to database.
    :param name: Name of interface.
    :param type_: Type of interface. 'loopback' or 'ethernet'.
    :param lladdr: Link layer address. Mostly MAC address.
    :return: Instance of added record or already existing record.
    """
    intf = ip_link_show(session, ifname=name)
    if intf:
        LOG.debug('Interface "%s" already exists: %s', intf.ifname, intf)
        return intf

    if type_ == 'ethernet':
        intf = Interface(
            ifname=name,
            flags=DEFAULT_ETH_FLAGS,
            ifmtu=DEFAULT_ETH_MTU,
            ifmtu6=DEFAULT_ETH_MTU,
            hw_addr=lladdr)
    else:  # type_ == 'loopback':
        intf = Interface(
            ifname=name,
            inet='127.0.0.1/8',
            inet6='::1/128')

    session.add(intf)

    return intf


@base.sql_function
def ip_link_delete(session, name):
    """
    Deletes an interface record from Zebra protocol service database.

    The arguments are similar to "ip link delete" command of iproute2.

    :param session: Session instance connecting to database.
    :param name: Name of interface.
    :return: Name of interface which was deleted. None if failed.
    """
    intf = ip_link_show(session, ifname=name)
    if not intf:
        LOG.debug('Interface "%s" does not exist', name)
        return None

    session.delete(intf)

    return name


# Currently, functions corresponding to "ip link show" and "ip address show"
# have the same implementation.
ip_address_show = ip_link_show
ip_address_show_all = ip_link_show_all


@base.sql_function
def ip_address_add(session, ifname, ifaddr):
    """
    Adds an IP address to interface record identified with the given "ifname".

    The arguments are similar to "ip address add" command of iproute2.

    :param session: Session instance connecting to database.
    :param ifname: Name of interface.
    :param ifaddr: IPv4 or IPv6 address.
    :return: Instance of record or "None" if failed.
    """
    def _append_inet_addr(intf_inet, addr):
        addr_list = intf_inet.split(',')
        if addr in addr_list:
            LOG.debug(
                'Interface "%s" has already "ifaddr": %s',
                intf.ifname, addr)
            return intf_inet
        else:
            addr_list.append(addr)
            return ','.join(addr_list)

    intf = ip_link_show(session, ifname=ifname)
    if not intf:
        LOG.debug('Interface "%s" does not exist', ifname)
        return None

    if ip.valid_ipv4(ifaddr):
        intf.inet = _append_inet_addr(intf.inet, ifaddr)
    elif ip.valid_ipv6(ifaddr):
        intf.inet6 = _append_inet_addr(intf.inet6, ifaddr)
    else:
        LOG.debug('Invalid IP address for "ifaddr": %s', ifaddr)
        return None

    return intf


@base.sql_function
def ip_address_delete(session, ifname, ifaddr):
    """
    Deletes an IP address from interface record identified with the given
    "ifname".

    The arguments are similar to "ip address delete" command of iproute2.

    :param session: Session instance connecting to database.
    :param ifname: Name of interface.
    :param ifaddr: IPv4 or IPv6 address.
    :return: Instance of record or "None" if failed.
    """
    def _remove_inet_addr(intf_inet, addr):
        addr_list = intf_inet.split(',')
        if addr not in addr_list:
            LOG.debug(
                'Interface "%s" does not have "ifaddr": %s',
                intf.ifname, addr)
            return intf_inet
        else:
            addr_list.remove(addr)
            return ','.join(addr_list)

    intf = ip_link_show(session, ifname=ifname)
    if not intf:
        LOG.debug('Interface "%s" does not exist', ifname)
        return None

    if ip.valid_ipv4(ifaddr):
        intf.inet = _remove_inet_addr(intf.inet, ifaddr)
    elif ip.valid_ipv6(ifaddr):
        intf.inet6 = _remove_inet_addr(intf.inet6, ifaddr)
    else:
        LOG.debug('Invalid IP address for "ifaddr": %s', ifaddr)
        return None

    return intf



# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Internal representation of port_no id is int(32bit)
# string representation is in hex without '0x'

_PORT_NO_LEN = 8
_PORT_NO_FMT = '%0{0}x'.format(_PORT_NO_LEN)
PORT_NO_PATTERN = r'[0-9a-f]{%d}' % _PORT_NO_LEN


def port_no_to_str(port_no):
    return _PORT_NO_FMT % port_no


def str_to_port_no(port_no_str):
    assert len(port_no_str) == _PORT_NO_LEN
    return int(port_no_str, 16)



# Copyright (C) 2014 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module provides a convenient application for using OSKen BGPSpeaker and for
writing your BGP application.

It reads a configuration file which includes settings for neighbors, routes
and some others.
Please refer to ``os_ken/services/protocols/bgp/bgp_sample_conf.py`` for the
sample configuration.

Usage Example::

    $ osken-manager os_ken/services/protocols/bgp/application.py \\
        --bgp-app-config-file os_ken/services/protocols/bgp/bgp_sample_conf.py

SSH Console
===========

You can also use the SSH console and see the RIB and do some operations from
this console.
The SSH port and username/password can be configured by the configuration file.
You can check the help by hitting '?' key in this interface.

Example::

    $ ssh localhost -p 4990

    Hello, this is OSKen BGP speaker (version 4.19).

    bgpd> # Hit '?' key
     clear - allows to reset BGP connections
     help - show this help
     quit - exit this session
     set - set runtime settings
     show - shows runtime state information
    bgpd>
    bgpd> show rib all
    Status codes: * valid, > best
    Origin codes: i - IGP, e - EGP, ? - incomplete
         Network        Labels   Next Hop   Reason      Metric LocPrf Path
     *>  10.10.1.0/24   None     0.0.0.0    Only Path                 i
    bgpd>

Integration with Other Applications
===================================

``os_ken.services.protocols.bgp.application.OSKenBGPSpeaker`` will notifies the
following events to other OSKen applications.

    - ``EventBestPathChanged``
    - ``EventAdjRibInChanged``
    - ``EventPeerDown``
    - ``EventPeerUp``

To catch these events, specify ``@set_ev_cls()`` decorator to the event
handlers in the OSKen applications.

Example Application::

    # my_bgp_app.py

    from os_ken.base import app_manager
    from os_ken.controller.handler import set_ev_cls
    from os_ken.services.protocols.bgp import application as bgp_application


    class MyBGPApp(app_manager.OSKenApp):
        _CONTEXTS = {
            'os_kenbgpspeaker': bgp_application.OSKenBGPSpeaker,
        }

        def __init__(self, *args, **kwargs):
            super(MyBGPApp, self).__init__(*args, **kwargs)

            # Stores "os_ken.services.protocols.bgp.application.OSKenBGPSpeaker"
            # instance in order to call the APIs of
            # "os_ken.services.protocols.bgp.bgpspeaker.BGPSpeaker" via
            # "self.app.speaker".
            # Please note at this time, "BGPSpeaker" is NOT instantiated yet.
            self.app = kwargs['os_kenbgpspeaker']

        @set_ev_cls(bgp_application.EventBestPathChanged)
        def _best_patch_changed_handler(self, ev):
            self.logger.info(
                'Best path changed: is_withdraw=%s, path=%s',
                ev.is_withdraw, ev.path)

Usage Example::

    $ osken-manager my_bgp_app.py \\
        --bgp-app-config-file os_ken/services/protocols/bgp/bgp_sample_conf.py

.. note::

    For the APIs for ``os_ken.services.protocols.bgp.bgpspeaker.BGPSpeaker``,
    please refer to :doc:`../library_bgp_speaker_ref`.

API Reference
=============
"""

import logging
import os

from os_ken import cfg
from os_ken.lib import hub
from os_ken.utils import load_source
from os_ken.base.app_manager import OSKenApp
from os_ken.controller.event import EventBase
from os_ken.services.protocols.bgp.base import add_bgp_error_metadata
from os_ken.services.protocols.bgp.base import BGPSException
from os_ken.services.protocols.bgp.base import BIN_ERROR
from os_ken.services.protocols.bgp.bgpspeaker import BGPSpeaker
from os_ken.services.protocols.bgp.net_ctrl import NET_CONTROLLER
from os_ken.services.protocols.bgp.net_ctrl import NC_RPC_BIND_IP
from os_ken.services.protocols.bgp.net_ctrl import NC_RPC_BIND_PORT
from os_ken.services.protocols.bgp.rtconf.base import RuntimeConfigError
from os_ken.services.protocols.bgp.rtconf.common import LOCAL_AS
from os_ken.services.protocols.bgp.rtconf.common import ROUTER_ID
from os_ken.services.protocols.bgp.utils.validation import is_valid_ipv4
from os_ken.services.protocols.bgp.utils.validation import is_valid_ipv6


LOG = logging.getLogger('bgpspeaker.application')

CONF = cfg.CONF['bgp-app']


@add_bgp_error_metadata(code=BIN_ERROR,
                        sub_code=1,
                        def_desc='Unknown bootstrap exception.')
class ApplicationException(BGPSException):
    """
    Specific Base exception related to `BSPSpeaker`.
    """
    pass


def validate_rpc_host(ip):
    """
    Validates the given ip for use as RPC server address.
    """
    if not is_valid_ipv4(ip) and not is_valid_ipv6(ip):
        raise ApplicationException(
            desc='Invalid RPC ip address: %s' % ip)
    return ip


def load_config(config_file):
    """
    Validates the given file for use as the settings file for BGPSpeaker
    and loads the configuration from the given file as a module instance.
    """
    if not config_file or not os.path.isfile(config_file):
        raise ApplicationException(
            desc='Invalid configuration file: %s' % config_file)

    # Loads the configuration from the given file, if available.
    try:
        return load_source('bgpspeaker.application.settings', config_file)
    except Exception as e:
        raise ApplicationException(desc=str(e))


class EventBestPathChanged(EventBase):
    """
    Event called when any best remote path is changed due to UPDATE messages
    or remote peer's down.

    This event is the wrapper for ``best_path_change_handler`` of
    ``bgpspeaker.BGPSpeaker``.

    ``path`` attribute contains an instance of ``info_base.base.Path``
    subclasses.

    If ``is_withdraw`` attribute is ``True``, ``path`` attribute has the
    information of the withdraw route.
    """

    def __init__(self, path, is_withdraw):
        super(EventBestPathChanged, self).__init__()
        self.path = path
        self.is_withdraw = is_withdraw


class EventAdjRibInChanged(EventBase):
    """
    Event called when any adj-RIB-in path is changed due to UPDATE messages
    or remote peer's down.

    This event is the wrapper for ``adj_rib_in_change_handler`` of
    ``bgpspeaker.BGPSpeaker``.

    ``path`` attribute contains an instance of ``info_base.base.Path``
    subclasses.

    If ``is_withdraw`` attribute is ``True``, ``path`` attribute has the
    information of the withdraw route.

    ``peer_ip`` is the peer's IP address who sent this path.

    ``peer_as`` is the peer's AS number who sent this path.
    """

    def __init__(self, path, is_withdraw, peer_ip, peer_as):
        super(EventAdjRibInChanged, self).__init__()
        self.path = path
        self.is_withdraw = is_withdraw
        self.peer_ip = peer_ip
        self.peer_as = peer_as


class EventPeerDown(EventBase):
    """
    Event called when the session to the remote peer goes down.

    This event is the wrapper for ``peer_down_handler`` of
    ``bgpspeaker.BGPSpeaker``.

    ``remote_ip`` attribute is the IP address of the remote peer.

    ``remote_as`` attribute is the AS number of the remote peer.
    """

    def __init__(self, remote_ip, remote_as):
        super(EventPeerDown, self).__init__()
        self.remote_ip = remote_ip
        self.remote_as = remote_as


class EventPeerUp(EventBase):
    """
    Event called when the session to the remote peer goes up.

    This event is the wrapper for ``peer_up_handler`` of
    ``bgpspeaker.BGPSpeaker``.

    ``remote_ip`` attribute is the IP address of the remote peer.

    ``remote_as`` attribute is the AS number of the remote peer.
    """

    def __init__(self, remote_ip, remote_as):
        super(EventPeerUp, self).__init__()
        self.remote_ip = remote_ip
        self.remote_as = remote_as


class OSKenBGPSpeaker(OSKenApp):
    """
    Base application for implementing BGP applications.
    """
    _EVENTS = [
        EventBestPathChanged,
        EventAdjRibInChanged,
        EventPeerDown,
        EventPeerUp,
    ]

    def __init__(self, *args, **kwargs):
        super(OSKenBGPSpeaker, self).__init__(*args, **kwargs)
        self.config_file = CONF.config_file

        # BGPSpeaker instance (not instantiated yet)
        self.speaker = None

    def start(self):
        super(OSKenBGPSpeaker, self).start()

        # If configuration file was provided and loaded successfully, we start
        # BGPSpeaker using the given settings.
        # If no configuration file is provided or if any minimum required
        # setting is missing, BGPSpeaker will not be started.
        if self.config_file:
            LOG.debug('Loading config file %s...', self.config_file)
            settings = load_config(self.config_file)

            # Configure logging settings, if available.
            if hasattr(settings, 'LOGGING'):
                # Not implemented yet.
                LOG.debug('Loading LOGGING settings... (NOT implemented yet)')
                # from logging.config import dictConfig
                # logging_settings = dictConfig(settings.LOGGING)

            # Configure BGP settings, if available.
            if hasattr(settings, 'BGP'):
                LOG.debug('Loading BGP settings...')
                self._start_speaker(settings.BGP)

            # Configure SSH settings, if available.
            if hasattr(settings, 'SSH'):
                LOG.debug('Loading SSH settings...')
                # Note: paramiko used in bgp.operator.ssh is the optional
                # requirements, imports bgp.operator.ssh here.
                from os_ken.services.protocols.bgp.operator import ssh
                hub.spawn(ssh.SSH_CLI_CONTROLLER.start, **settings.SSH)

        # Start RPC server with the given RPC settings.
        rpc_settings = {
            NC_RPC_BIND_PORT: CONF.rpc_port,
            NC_RPC_BIND_IP: validate_rpc_host(CONF.rpc_host),
        }
        return hub.spawn(NET_CONTROLLER.start, **rpc_settings)

    def _start_speaker(self, settings):
        """
        Starts BGPSpeaker using the given settings.
        """
        # Check required settings.
        _required_settings = (
            LOCAL_AS,
            ROUTER_ID,
        )
        for required in _required_settings:
            if required not in settings:
                raise ApplicationException(
                    desc='Required BGP configuration missing: %s' % required)

        # Set event notify handlers if no corresponding handler specified.
        settings.setdefault(
            'best_path_change_handler', self._notify_best_path_changed_event)
        settings.setdefault(
            'adj_rib_in_change_handler', self._notify_adj_rib_in_changed_event)
        settings.setdefault(
            'peer_down_handler', self._notify_peer_down_event)
        settings.setdefault(
            'peer_up_handler', self._notify_peer_up_event)

        # Pop settings other than creating BGPSpeaker instance.
        neighbors_settings = settings.pop('neighbors', [])
        vrfs_settings = settings.pop('vrfs', [])
        routes_settings = settings.pop('routes', [])

        # Create BGPSpeaker instance.
        LOG.debug('Starting BGPSpeaker...')
        settings.setdefault('as_number', settings.pop(LOCAL_AS))
        self.speaker = BGPSpeaker(**settings)

        # Add neighbors.
        LOG.debug('Adding neighbors...')
        self._add_neighbors(neighbors_settings)

        # Add VRFs.
        LOG.debug('Adding VRFs...')
        self._add_vrfs(vrfs_settings)

        # Add routes
        LOG.debug('Adding routes...')
        self._add_routes(routes_settings)

    def _notify_best_path_changed_event(self, ev):
        ev = EventBestPathChanged(ev.path, ev.is_withdraw)
        self.send_event_to_observers(ev)

    def _notify_adj_rib_in_changed_event(self, ev, peer_ip, peer_as):
        ev = EventAdjRibInChanged(ev.path, ev.is_withdraw, peer_ip, peer_as)
        self.send_event_to_observers(ev)

    def _notify_peer_down_event(self, remote_ip, remote_as):
        ev = EventPeerDown(remote_ip, remote_as)
        self.send_event_to_observers(ev)

    def _notify_peer_up_event(self, remote_ip, remote_as):
        ev = EventPeerUp(remote_ip, remote_as)
        self.send_event_to_observers(ev)

    def _add_neighbors(self, settings):
        """
        Add BGP neighbors from the given settings.

        All valid neighbors are loaded.
        Miss-configured neighbors are ignored and errors are logged.
        """
        for neighbor_settings in settings:
            LOG.debug('Adding neighbor settings: %s', neighbor_settings)
            try:
                self.speaker.neighbor_add(**neighbor_settings)
            except RuntimeConfigError as e:
                LOG.exception(e)

    def _add_vrfs(self, settings):
        """
        Add BGP VRFs from the given settings.

        All valid VRFs are loaded.
        Miss-configured VRFs are ignored and errors are logged.
        """
        for vrf_settings in settings:
            LOG.debug('Adding VRF settings: %s', vrf_settings)
            try:
                self.speaker.vrf_add(**vrf_settings)
            except RuntimeConfigError as e:
                LOG.exception(e)

    def _add_routes(self, settings):
        """
        Add BGP routes from given settings.

        All valid routes are loaded.
        Miss-configured routes are ignored and errors are logged.
        """
        for route_settings in settings:
            if 'prefix' in route_settings:
                prefix_add = self.speaker.prefix_add
            elif 'route_type' in route_settings:
                prefix_add = self.speaker.evpn_prefix_add
            elif 'flowspec_family' in route_settings:
                prefix_add = self.speaker.flowspec_prefix_add
            else:
                LOG.debug('Skip invalid route settings: %s', route_settings)
                continue

            LOG.debug('Adding route settings: %s', route_settings)
            try:
                prefix_add(**route_settings)
            except RuntimeConfigError as e:
                LOG.exception(e)



# Copyright (C) 2014 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
 Defines some model classes related BGP.

 These class include types used in saving information sent/received over BGP
 sessions.
"""
import logging
from time import gmtime


LOG = logging.getLogger('bgpspeaker.model')


class Counter(object):
    """Simple counter for keeping count of several keys."""

    def __init__(self):
        self._counters = {}

    def incr(self, counter_name, incr_by=1):
        self._counters[counter_name] = \
            self._counters.get(counter_name, 0) + incr_by

    def get_count(self, counter_name):
        return self._counters.get(counter_name, 0)

    def get_counters(self):
        return self._counters.copy()


class OutgoingRoute(object):
    """Holds state about a route that is queued for being sent to a given sink.
    """

    __slots__ = ('_path', '_for_route_refresh',
                 'sink', 'next_outgoing_route', 'prev_outgoing_route',
                 'next_sink_out_route', 'prev_sink_out_route')

    def __init__(self, path, for_route_refresh=False):
        assert path

        self.sink = None

        self._path = path

        # Is this update in response for route-refresh request.
        # No sent-route is queued for the destination for this update.
        self._for_route_refresh = for_route_refresh

        # Automatically generated, for list off of Destination.
        #
        # self.next_outgoing_route
        # self.prev_outgoing_route

        # Automatically generated for list off of sink.
        #
        # self.next_sink_out_route
        # self.prev_sink_out_route

    @property
    def path(self):
        return self._path

    @property
    def for_route_refresh(self):
        return self._for_route_refresh

    def __str__(self):
        return ('OutgoingRoute(path: %s, for_route_refresh: %s)' %
                (self.path, self.for_route_refresh))


class FlexinetOutgoingRoute(object):
    """Holds state about a route that is queued for being sent to a given sink.

    In this case the sink is flexinet peer and this route information is from
    a VRF which holds Ipv4(v6) NLRIs.
    """

    __slots__ = ('_path', 'sink', 'next_outgoing_route', 'prev_outgoing_route',
                 'next_sink_out_route', 'prev_sink_out_route', '_route_dist')

    def __init__(self, path, route_dist):
        from os_ken.services.protocols.bgp.info_base.vrf4 import Vrf4Path
        from os_ken.services.protocols.bgp.info_base.vrf6 import Vrf6Path
        from os_ken.services.protocols.bgp.info_base.vrfevpn import VrfEvpnPath
        from os_ken.services.protocols.bgp.info_base.vrf4fs import (
            Vrf4FlowSpecPath)
        from os_ken.services.protocols.bgp.info_base.vrf6fs import (
            Vrf6FlowSpecPath)
        from os_ken.services.protocols.bgp.info_base.vrfl2vpnfs import (
            L2vpnFlowSpecPath)
        assert path.route_family in (Vrf4Path.ROUTE_FAMILY,
                                     Vrf6Path.ROUTE_FAMILY,
                                     VrfEvpnPath.ROUTE_FAMILY,
                                     Vrf4FlowSpecPath.ROUTE_FAMILY,
                                     Vrf6FlowSpecPath.ROUTE_FAMILY,
                                     L2vpnFlowSpecPath.ROUTE_FAMILY,
                                     )

        self.sink = None
        self._path = path
        self._route_dist = route_dist

        # Automatically generated, for list off of Destination.
        #
        # self.next_outgoing_route
        # self.prev_outgoing_route

        # Automatically generated for list off of sink.
        #
        # self.next_sink_out_route
        # self.prev_sink_out_route

    @property
    def path(self):
        return self._path

    @property
    def route_dist(self):
        return self._route_dist

    def __str__(self):
        return ('FlexinetOutgoingRoute(path: %s, route_dist: %s)' %
                (self.path, self.route_dist))


class SentRoute(object):
    """Holds the information that has been sent to one or more sinks
    about a particular BGP destination.
    """

    def __init__(self, path, peer, filtered=None, timestamp=None):
        assert (path and hasattr(peer, 'version_num'))

        self.path = path

        # Peer to which this path was sent.
        self._sent_peer = peer

        self.filtered = filtered

        if timestamp:
            self.timestamp = timestamp
        else:
            self.timestamp = gmtime()

        # Automatically generated.
        #
        # self.next_sent_route
        # self.prev_sent_route

    @property
    def sent_peer(self):
        return self._sent_peer


class ReceivedRoute(object):
    """Holds the information that has been received to one sinks
    about a particular BGP destination.
    """

    def __init__(self, path, peer, filtered=None, timestamp=None):
        assert (path and hasattr(peer, 'version_num'))

        self.path = path

        # Peer to which this path was received.
        self._received_peer = peer

        self.filtered = filtered

        if timestamp:
            self.timestamp = timestamp
        else:
            self.timestamp = gmtime()

    @property
    def received_peer(self):
        return self._received_peer



# Copyright (C) 2014 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
 Defines classes related to incorrect parameters.
"""
from os_ken.services.protocols.bgp.operator.command import CommandsResponse
from os_ken.services.protocols.bgp.operator.command import STATUS_ERROR
from os_ken.services.protocols.bgp.operator.internal_api import WrongParamError


class WrongParamResp(object):
    def __new__(cls, e=None):
        return cls.wrong_param_resp_factory(e)

    @staticmethod
    def wrong_param_resp_factory(e=None):
        if not e:
            e = WrongParamError()
        desc = 'wrong parameters: %s' % str(e)

        return CommandsResponse(STATUS_ERROR, desc)



# Copyright (C) 2014 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
 Defined log handler to be used to log to RPC connection.
"""

import logging

from os_ken.services.protocols.bgp.net_ctrl import NET_CONTROLLER
from os_ken.services.protocols.bgp.net_ctrl import NOTIFICATION_LOG


class RpcLogHandler(logging.Handler):
    """Outputs log records to `NET_CONTROLLER`."""

    def emit(self, record):
        msg = self.format(record)
        NET_CONTROLLER.send_rpc_notification(
            NOTIFICATION_LOG,
            {
                'level': record.levelname,
                'msg': msg
            }
        )



# Copyright (C) 2012 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# vim: tabstop=4 shiftwidth=4 softtabstop=4

import unittest
import logging
from os_ken.ofproto.inet import *


LOG = logging.getLogger('test_inet')


class TestInet(unittest.TestCase):
    """ Test case for inet
    """

    def test_ip_proto(self):
        self.assertEqual(IPPROTO_IP, 0)
        self.assertEqual(IPPROTO_HOPOPTS, 0)
        self.assertEqual(IPPROTO_ICMP, 1)
        self.assertEqual(IPPROTO_TCP, 6)
        self.assertEqual(IPPROTO_UDP, 17)
        self.assertEqual(IPPROTO_ROUTING, 43)
        self.assertEqual(IPPROTO_FRAGMENT, 44)
        self.assertEqual(IPPROTO_AH, 51)
        self.assertEqual(IPPROTO_ICMPV6, 58)
        self.assertEqual(IPPROTO_NONE, 59)
        self.assertEqual(IPPROTO_DSTOPTS, 60)
        self.assertEqual(IPPROTO_SCTP, 132)



# Copyright (C) 2012 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct

from . import packet_base
from os_ken.lib import type_desc


class mpls(packet_base.PacketBase):
    """MPLS (RFC 3032) header encoder/decoder class.

    NOTE: When decoding, this implementation assumes that the inner protocol
    is IPv4.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    ============== ====================
    Attribute      Description
    ============== ====================
    label          Label Value
    exp            Experimental Use
    bsb            Bottom of Stack
    ttl            Time To Live
    ============== ====================
    """

    _PACK_STR = '!I'
    _MIN_LEN = struct.calcsize(_PACK_STR)

    def __init__(self, label=0, exp=0, bsb=1, ttl=255):
        super(mpls, self).__init__()
        self.label = label
        self.exp = exp
        self.bsb = bsb
        self.ttl = ttl

    @classmethod
    def parser(cls, buf):
        (label,) = struct.unpack_from(cls._PACK_STR, buf)
        ttl = label & 0xff
        bsb = (label >> 8) & 1
        exp = (label >> 9) & 7
        label = label >> 12
        msg = cls(label, exp, bsb, ttl)
        if bsb:
            from . import ipv4
            return msg, ipv4.ipv4, buf[msg._MIN_LEN:]
        else:
            return msg, mpls, buf[msg._MIN_LEN:]

    def serialize(self, payload, prev):
        val = self.label << 12 | self.exp << 9 | self.bsb << 8 | self.ttl
        return struct.pack(mpls._PACK_STR, val)


def label_from_bin(buf):
    """
    Converts binary representation label to integer.

    :param buf: Binary representation of label.
    :return: MPLS Label and BoS bit.
    """

    mpls_label = type_desc.Int3.to_user(bytes(buf))
    return mpls_label >> 4, mpls_label & 1


def label_to_bin(mpls_label, is_bos=True):
    """
    Converts integer label to binary representation.

    :param mpls_label: MPLS Label.
    :param is_bos: BoS bit.
    :return: Binary representation of label.
    """
    return type_desc.Int3.from_user(mpls_label << 4 | is_bos)



# Copyright (C) 2012 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# vim: tabstop=4 shiftwidth=4 softtabstop=4

import sys
import logging
import itertools

from os_ken import utils
from os_ken.lib import mac
from os_ken.base import app_manager
from os_ken.controller import ofp_event
from os_ken.controller import handler
from os_ken.controller import dpset
from os_ken.controller.handler import MAIN_DISPATCHER
from os_ken.controller.handler import CONFIG_DISPATCHER
from os_ken.controller.handler import set_ev_cls
from os_ken.ofproto import ofproto_v1_0
from os_ken.ofproto import ofproto_v1_2


LOG = logging.getLogger(__name__)


LOG_TEST_START = 'TEST_START: %s'
LOG_TEST_RESULTS = 'TEST_RESULTS:'
LOG_TEST_FINISH = 'TEST_FINISHED: Completed=[%s] (OK=%s NG=%s SKIP=%s)'
LOG_TEST_UNSUPPORTED = 'SKIP (unsupported)'


class TestFlowBase(app_manager.OSKenApp):
    """
    To run the tests is required for the following pair of functions.
        1. test_<test name>()
            To send flows to switch.

        2. verify_<test name>() or _verify_default()
            To check flows of switch.
    """

    _CONTEXTS = {'dpset': dpset.DPSet}

    def __init__(self, *args, **kwargs):
        super(TestFlowBase, self).__init__(*args, **kwargs)
        self.pending = []
        self.results = {}
        self.current = None
        self.unclear = 0

        for t in dir(self):
            if t.startswith("test_"):
                self.pending.append(t)
        self.pending.sort(reverse=True)
        self.unclear = len(self.pending)

    def delete_all_flows(self, dp):
        if dp.ofproto == ofproto_v1_0:
            match = dp.ofproto_parser.OFPMatch(dp.ofproto.OFPFW_ALL,
                                               0, 0, 0, 0, 0,
                                               0, 0, 0, 0, 0, 0, 0)
            m = dp.ofproto_parser.OFPFlowMod(dp, match, 0,
                                             dp.ofproto.OFPFC_DELETE,
                                             0, 0, 0, 0,
                                             dp.ofproto.OFPP_NONE, 0, None)
        elif dp.ofproto == ofproto_v1_2:
            match = dp.ofproto_parser.OFPMatch()
            m = dp.ofproto_parser.OFPFlowMod(dp, 0, 0, dp.ofproto.OFPTT_ALL,
                                             dp.ofproto.OFPFC_DELETE,
                                             0, 0, 0, 0xffffffff,
                                             dp.ofproto.OFPP_ANY,
                                             dp.ofproto.OFPG_ANY,
                                             0, match, [])

        dp.send_msg(m)

    def send_flow_stats(self, dp):
        if dp.ofproto == ofproto_v1_0:
            match = dp.ofproto_parser.OFPMatch(dp.ofproto.OFPFW_ALL,
                                               0, 0, 0, 0, 0,
                                               0, 0, 0, 0, 0, 0, 0)
            m = dp.ofproto_parser.OFPFlowStatsRequest(dp, 0, match,
                                                      0, dp.ofproto.OFPP_NONE)
        elif dp.ofproto == ofproto_v1_2:
            match = dp.ofproto_parser.OFPMatch()
            m = dp.ofproto_parser.OFPFlowStatsRequest(dp, dp.ofproto.OFPTT_ALL,
                                                      dp.ofproto.OFPP_ANY,
                                                      dp.ofproto.OFPG_ANY,
                                                      0, 0, match)

        dp.send_msg(m)

    def verify_default(self, dp, stats):
        return 'function %s() is not found.' % ("verify" + self.current[4:], )

    def start_next_test(self, dp):
        self.delete_all_flows(dp)
        dp.send_barrier()
        if len(self.pending):
            t = self.pending.pop()
            if self.is_supported(t):
                LOG.info(LOG_TEST_START, t)
                self.current = t
                getattr(self, t)(dp)
                dp.send_barrier()
                self.send_flow_stats(dp)
            else:
                self.results[t] = LOG_TEST_UNSUPPORTED
                self.unclear -= 1
                self.start_next_test(dp)
        else:
            self.print_results()

    def print_results(self):
        LOG.info("TEST_RESULTS:")
        ok = 0
        ng = 0
        skip = 0
        for t in sorted(self.results.keys()):
            if self.results[t] is True:
                ok += 1
            elif self.results[t] == LOG_TEST_UNSUPPORTED:
                skip += 1
            else:
                ng += 1
            LOG.info("    %s: %s", t, self.results[t])
        LOG.info(LOG_TEST_FINISH, self.unclear == 0, ok, ng, skip)

    @handler.set_ev_cls(ofp_event.EventOFPFlowStatsReply,
                        handler.MAIN_DISPATCHER)
    def flow_reply_handler(self, ev):
        self.run_verify(ev)

    @handler.set_ev_cls(ofp_event.EventOFPStatsReply,
                        handler.MAIN_DISPATCHER)
    def stats_reply_handler(self, ev):
        self.run_verify(ev)

    def run_verify(self, ev):
        msg = ev.msg
        dp = msg.datapath

        verify_func = self.verify_default
        v = "verify" + self.current[4:]
        if hasattr(self, v):
            verify_func = getattr(self, v)

        result = verify_func(dp, msg.body)
        if result is True:
            self.unclear -= 1

        self.results[self.current] = result
        self.start_next_test(dp)

    @handler.set_ev_cls(dpset.EventDP)
    def handler_datapath(self, ev):
        if ev.enter:
            self.start_next_test(ev.dp)

    @set_ev_cls(ofp_event.EventOFPBarrierReply, MAIN_DISPATCHER)
    def barrier_replay_handler(self, ev):
        pass

    def haddr_to_str(self, addr):
        return mac.haddr_to_str(addr)

    def haddr_to_bin(self, string):
        return mac.haddr_to_bin(string)

    def haddr_masked(self, haddr_bin, mask_bin):
        return mac.haddr_bitand(haddr_bin, mask_bin)

    def ipv4_to_str(self, integre):
        ip_list = [str((integre >> (24 - (n * 8)) & 255)) for n in range(4)]
        return '.'.join(ip_list)

    def ipv4_to_int(self, string):
        ip = string.split('.')
        assert len(ip) == 4
        i = 0
        for b in ip:
            b = int(b)
            i = (i << 8) | b
        return i

    def ipv4_masked(self, ip_int, mask_int):
        return ip_int & mask_int

    def ipv6_to_str(self, integres):
        return ':'.join(hex(x)[2:] for x in integres)

    def ipv6_to_int(self, string):
        ip = string.split(':')
        assert len(ip) == 8
        return [int(x, 16) for x in ip]

    def ipv6_masked(self, ipv6_int, mask_int):
        return [x & y for (x, y) in
                itertools.izip(ipv6_int, mask_int)]

    def is_supported(self, t):
        return True



# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import inspect
import logging
import struct
import unittest

from os_ken.lib import addrconv
from os_ken.lib.packet import dhcp


LOG = logging.getLogger(__name__)


class Test_dhcp_offer(unittest.TestCase):

    op = dhcp.DHCP_BOOT_REPLY
    chaddr = 'aa:aa:aa:aa:aa:aa'
    htype = 1
    hlen = 6
    hops = 0
    xid = 1
    secs = 0
    flags = 1
    ciaddr = '192.168.10.10'
    yiaddr = '192.168.20.20'
    siaddr = '192.168.30.30'
    giaddr = '192.168.40.40'
    sname = 'abc'
    boot_file = ''

    option_list = [
        dhcp.option(dhcp.DHCP_MESSAGE_TYPE_OPT, b'\x02', 1),
        dhcp.option(dhcp.DHCP_SUBNET_MASK_OPT, b'\xff\xff\xff\x00', 4),
        dhcp.option(dhcp.DHCP_GATEWAY_ADDR_OPT, b'\xc0\xa8\x0a\x09', 4),
        dhcp.option(dhcp.DHCP_DNS_SERVER_ADDR_OPT, b'\xc0\xa8\x0a\x09', 4),
        dhcp.option(dhcp.DHCP_IP_ADDR_LEASE_TIME_OPT, b'\x00\x03\xf4\x80', 4),
        dhcp.option(dhcp.DHCP_RENEWAL_TIME_OPT, b'\x00\x01\xfa\x40', 4),
        dhcp.option(dhcp.DHCP_REBINDING_TIME_OPT, b'\x00\x03\x75\xf0', 4),
        dhcp.option(dhcp.DHCP_SERVER_IDENTIFIER_OPT, b'\xc0\xa8\x0a\x09', 4)]
    magic_cookie = '99.130.83.99'
    options = dhcp.options(option_list=option_list, options_len=50,
                           magic_cookie=magic_cookie)

    dh = dhcp.dhcp(op, chaddr, options, htype=htype, hlen=hlen,
                   hops=hops, xid=xid, secs=secs, flags=flags,
                   ciaddr=ciaddr, yiaddr=yiaddr, siaddr=siaddr,
                   giaddr=giaddr, sname=sname, boot_file=boot_file)

    buf = (
        b"\x02\x01\x06\x00\x00\x00\x00\x01\x00\x00\x00\x01\xc0\xa8\x0a\x0a"
        b"\xc0\xa8\x14\x14\xc0\xa8\x1e\x1e\xc0\xa8\x28\x28\xaa\xaa\xaa\xaa"
        b"\xaa\xaa\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x61\x62\x63\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x63\x82\x53\x63"
        b"\x35\x01\x02\x01\x04\xff\xff\xff\x00\x03\x04\xc0\xa8\x0a\x09\x06"
        b"\x04\xc0\xa8\x0a\x09\x33\x04\x00\x03\xf4\x80\x3a\x04\x00\x01\xfa"
        b"\x40\x3b\x04\x00\x03\x75\xf0\x36\x04\xc0\xa8\x0a\x09\xff")

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_init(self):
        self.assertEqual(self.op, self.dh.op)
        self.assertEqual(self.htype, self.dh.htype)
        self.assertEqual(self.hlen, self.dh.hlen)
        self.assertEqual(self.hops, self.dh.hops)
        self.assertEqual(self.xid, self.dh.xid)
        self.assertEqual(self.secs, self.dh.secs)
        self.assertEqual(self.flags, self.dh.flags)
        self.assertEqual(self.ciaddr, self.dh.ciaddr)
        self.assertEqual(self.yiaddr, self.dh.yiaddr)
        self.assertEqual(self.siaddr, self.dh.siaddr)
        self.assertEqual(self.giaddr, self.dh.giaddr)
        self.assertEqual(self.chaddr, self.dh.chaddr)
        self.assertEqual(self.sname, self.dh.sname)
        self.assertEqual(self.boot_file, self.dh.boot_file)
        self.assertEqual(str(self.options), str(self.dh.options))

    def test_parser(self):
        res, _, rest = dhcp.dhcp.parser(self.buf)

        self.assertEqual(self.op, res.op)
        self.assertEqual(self.htype, res.htype)
        self.assertEqual(self.hlen, res.hlen)
        self.assertEqual(self.hops, res.hops)
        self.assertEqual(self.xid, res.xid)
        self.assertEqual(self.secs, res.secs)
        self.assertEqual(self.flags, res.flags)
        self.assertEqual(self.ciaddr, res.ciaddr)
        self.assertEqual(self.yiaddr, res.yiaddr)
        self.assertEqual(self.siaddr, res.siaddr)
        self.assertEqual(self.giaddr, res.giaddr)
        self.assertEqual(self.chaddr, res.chaddr)
        # sname is 64 byte length. rest of data is filled by '\x00'.
        self.assertEqual(self.sname.ljust(64, '\x00'), res.sname)
        # boof_file is 128 byte length. rest of data is filled by '\x00'.
        self.assertEqual(self.boot_file.ljust(128, '\x00'), res.boot_file)
        self.assertEqual(str(self.options), str(res.options))
        self.assertEqual(b'', rest)

    def test_parser_corrupted(self):
        corrupt_buf = self.buf[:-4]
        pkt, _, rest = dhcp.dhcp.parser(corrupt_buf)

        self.assertIsInstance(pkt, dhcp.dhcp)
        self.assertIsInstance(pkt.options, dhcp.options)
        for opt in pkt.options.option_list[:-1]:
            self.assertIsInstance(opt, dhcp.option)
        self.assertIsInstance(pkt.options.option_list[-1], bytes)

        buf = pkt.serialize()
        self.assertEqual(str(buf), str(corrupt_buf))
        self.assertEqual(b'', rest)

    def test_serialize(self):
        buf = self.dh.serialize()

        res = struct.unpack_from(dhcp.dhcp._DHCP_PACK_STR,
                                 bytes(buf))

        self.assertEqual(self.op, res[0])
        self.assertEqual(self.htype, res[1])
        self.assertEqual(self.hlen, res[2])
        self.assertEqual(self.hops, res[3])
        self.assertEqual(self.xid, res[4])
        self.assertEqual(self.secs, res[5])
        self.assertEqual(self.flags, res[6])
        self.assertEqual(self.ciaddr, addrconv.ipv4.bin_to_text(res[7]))
        self.assertEqual(self.yiaddr, addrconv.ipv4.bin_to_text(res[8]))
        self.assertEqual(self.siaddr, addrconv.ipv4.bin_to_text(res[9]))
        self.assertEqual(self.giaddr, addrconv.ipv4.bin_to_text(res[10]))
        self.assertEqual(self.chaddr, addrconv.mac.bin_to_text(res[11][:6]))
        # sname is 64 byte length. rest of data is filled by '\x00'.
        self.assertEqual(self.sname.ljust(64, '\x00'), res[12].decode('ascii'))
        # boof_file is 128 byte length. rest of data is filled by '\x00'.
        self.assertEqual(self.boot_file.ljust(128, '\x00'), res[13].decode('ascii'))
        options = dhcp.options.parser(
            buf[struct.calcsize(dhcp.dhcp._DHCP_PACK_STR):])
        self.assertEqual(str(self.options), str(options))

    def test_to_string(self):
        option_values = ['tag', 'length', 'value']
        opt_str_list = []
        for option in self.option_list:
            _opt_str = ','.join(['%s=%s' % (k, repr(getattr(option, k)))
                                 for k, v in inspect.getmembers(option)
                                 if k in option_values])
            opt_str = '%s(%s)' % (dhcp.option.__name__, _opt_str)
            opt_str_list.append(opt_str)
        option_str = '[%s]' % ', '.join(opt_str_list)

        opts_vals = {'magic_cookie': repr(self.magic_cookie),
                     'option_list': option_str,
                     'options_len': repr(self.options.options_len)}
        _options_str = ','.join(['%s=%s' % (k, opts_vals[k])
                                 for k, v in inspect.getmembers(self.options)
                                 if k in opts_vals])
        options_str = '%s(%s)' % (dhcp.options.__name__, _options_str)

        dhcp_values = {'op': repr(self.op),
                       'htype': repr(self.htype),
                       'hlen': repr(self.hlen),
                       'hops': repr(self.hops),
                       'xid': repr(self.xid),
                       'secs': repr(self.secs),
                       'flags': repr(self.flags),
                       'ciaddr': repr(self.ciaddr),
                       'yiaddr': repr(self.yiaddr),
                       'siaddr': repr(self.siaddr),
                       'giaddr': repr(self.giaddr),
                       'chaddr': repr(self.chaddr),
                       'sname': repr(self.sname),
                       'boot_file': repr(self.boot_file),
                       'options': options_str}
        _dh_str = ','.join(['%s=%s' % (k, dhcp_values[k])
                            for k, v in inspect.getmembers(self.dh)
                            if k in dhcp_values])
        dh_str = '%s(%s)' % (dhcp.dhcp.__name__, _dh_str)

        self.assertEqual(str(self.dh), dh_str)
        self.assertEqual(repr(self.dh), dh_str)

    def test_json(self):
        jsondict = self.dh.to_jsondict()
        dh = dhcp.dhcp.from_jsondict(jsondict['dhcp'])
        self.assertEqual(str(self.dh), str(dh))



# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import logging
import struct
import unittest

from os_ken.lib import addrconv
from os_ken.lib.packet import packet
from os_ken.lib.packet import ethernet
from os_ken.lib.packet import ipv4
from os_ken.lib.packet import sctp
from os_ken.ofproto import ether
from os_ken.ofproto import inet


LOG = logging.getLogger(__name__)


class Test_sctp(unittest.TestCase):

    def setUp(self):
        self.chunks = []
        self.csum = 0
        self.dst_port = 1234
        self.src_port = 5678
        self.vtag = 98765432

        self.sc = sctp.sctp(
            self.src_port, self.dst_port, self.vtag, self.csum,
            self.chunks)

        self.buf = b'\x16\x2e\x04\xd2\x05\xe3\x0a\x78\x00\x00\x00\x00'

    def setUp_with_data(self):
        self.unordered = 1
        self.begin = 1
        self.end = 1
        self.length = 16 + 10
        self.tsn = 12345
        self.sid = 1
        self.seq = 0
        self.payload_id = 0
        self.payload_data = b'\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a'

        self.data = sctp.chunk_data(
            unordered=self.unordered, begin=self.begin, end=self.end,
            tsn=self.tsn, sid=self.sid, payload_data=self.payload_data)

        self.chunks = [self.data]

        self.sc = sctp.sctp(
            self.src_port, self.dst_port, self.vtag, self.csum,
            self.chunks)

        self.buf += b'\x00\x07\x00\x1a\x00\x00\x30\x39\x00\x01\x00\x00' + \
            b'\x00\x00\x00\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a'

    def setUp_with_init(self):
        self.flags = 0
        self.length = 20 + 8 + 20 + 8 + 4 + 16 + 16
        self.init_tag = 123456
        self.a_rwnd = 9876
        self.os = 3
        self.mis = 3
        self.i_tsn = 123456

        self.p_ipv4 = sctp.param_ipv4('192.168.1.1')
        self.p_ipv6 = sctp.param_ipv6('fe80::647e:1aff:fec4:8284')
        self.p_cookie_preserve = sctp.param_cookie_preserve(5000)
        self.p_ecn = sctp.param_ecn()
        self.p_host_addr = sctp.param_host_addr(b'test host\x00')
        self.p_support_type = sctp.param_supported_addr(
            [sctp.PTYPE_IPV4, sctp.PTYPE_IPV6, sctp.PTYPE_COOKIE_PRESERVE,
             sctp.PTYPE_ECN, sctp.PTYPE_HOST_ADDR])
        self.params = [
            self.p_ipv4, self.p_ipv6, self.p_cookie_preserve,
            self.p_ecn, self.p_host_addr, self.p_support_type]

        self.init = sctp.chunk_init(
            init_tag=self.init_tag, a_rwnd=self.a_rwnd, os=self.os,
            mis=self.mis, i_tsn=self.i_tsn, params=self.params)

        self.chunks = [self.init]

        self.sc = sctp.sctp(
            self.src_port, self.dst_port, self.vtag, self.csum,
            self.chunks)

        self.buf += b'\x01\x00\x00\x5c\x00\x01\xe2\x40\x00\x00\x26\x94' + \
            b'\x00\x03\x00\x03\x00\x01\xe2\x40' + \
            b'\x00\x05\x00\x08\xc0\xa8\x01\x01' + \
            b'\x00\x06\x00\x14' + \
            b'\xfe\x80\x00\x00\x00\x00\x00\x00' + \
            b'\x64\x7e\x1a\xff\xfe\xc4\x82\x84' + \
            b'\x00\x09\x00\x08\x00\x00\x13\x88' + \
            b'\x80\x00\x00\x04' + \
            b'\x00\x0b\x00\x0e' + \
            b'\x74\x65\x73\x74\x20\x68\x6f\x73\x74\x00\x00\x00' + \
            b'\x00\x0c\x00\x0e\x00\x05\x00\x06\x00\x09\x80\x00' + \
            b'\x00\x0b\x00\x00'

    def setUp_with_init_ack(self):
        self.flags = 0
        self.length = 20 + 8 + 8 + 20 + 8 + 4 + 16
        self.init_tag = 123456
        self.a_rwnd = 9876
        self.os = 3
        self.mis = 3
        self.i_tsn = 123456

        self.p_state_cookie = sctp.param_state_cookie(b'\x01\x02\x03')
        self.p_ipv4 = sctp.param_ipv4('192.168.1.1')
        self.p_ipv6 = sctp.param_ipv6('fe80::647e:1aff:fec4:8284')
        self.p_unrecognized_param = sctp.param_unrecognized_param(
            b'\xff\xff\x00\x04')
        self.p_ecn = sctp.param_ecn()
        self.p_host_addr = sctp.param_host_addr(b'test host\x00')
        self.params = [
            self.p_state_cookie, self.p_ipv4, self.p_ipv6,
            self.p_unrecognized_param, self.p_ecn, self.p_host_addr]

        self.init_ack = sctp.chunk_init_ack(
            init_tag=self.init_tag, a_rwnd=self.a_rwnd, os=self.os,
            mis=self.mis, i_tsn=self.i_tsn, params=self.params)

        self.chunks = [self.init_ack]

        self.sc = sctp.sctp(
            self.src_port, self.dst_port, self.vtag, self.csum,
            self.chunks)

        self.buf += b'\x02\x00\x00\x54\x00\x01\xe2\x40\x00\x00\x26\x94' + \
            b'\x00\x03\x00\x03\x00\x01\xe2\x40' + \
            b'\x00\x07\x00\x07\x01\x02\x03\x00' + \
            b'\x00\x05\x00\x08\xc0\xa8\x01\x01' + \
            b'\x00\x06\x00\x14' + \
            b'\xfe\x80\x00\x00\x00\x00\x00\x00' + \
            b'\x64\x7e\x1a\xff\xfe\xc4\x82\x84' + \
            b'\x00\x08\x00\x08\xff\xff\x00\x04' + \
            b'\x80\x00\x00\x04' + \
            b'\x00\x0b\x00\x0e' + \
            b'\x74\x65\x73\x74\x20\x68\x6f\x73\x74\x00\x00\x00'

    def setUp_with_sack(self):
        self.flags = 0
        self.length = 16 + 2 * 2 * 5 + 4 * 5
        self.tsn_ack = 123456
        self.a_rwnd = 9876
        self.gapack_num = 5
        self.duptsn_num = 5
        self.gapacks = [[2, 3], [10, 12], [20, 24], [51, 52], [62, 63]]
        self.duptsns = [123458, 123466, 123476, 123507, 123518]

        self.sack = sctp.chunk_sack(
            tsn_ack=self.tsn_ack, a_rwnd=self.a_rwnd,
            gapack_num=self.gapack_num, duptsn_num=self.duptsn_num,
            gapacks=self.gapacks, duptsns=self.duptsns)

        self.chunks = [self.sack]

        self.sc = sctp.sctp(
            self.src_port, self.dst_port, self.vtag, self.csum,
            self.chunks)

        self.buf += b'\x03\x00\x00\x38\x00\x01\xe2\x40' + \
            b'\x00\x00\x26\x94\x00\x05\x00\x05' + \
            b'\x00\x02\x00\x03\x00\x0a\x00\x0c\x00\x14\x00\x18' + \
            b'\x00\x33\x00\x34\x00\x3e\x00\x3f' + \
            b'\x00\x01\xe2\x42\x00\x01\xe2\x4a\x00\x01\xe2\x54' + \
            b'\x00\x01\xe2\x73\x00\x01\xe2\x7e'

    def setUp_with_heartbeat(self):
        self.flags = 0
        self.length = 4 + 8

        self.p_heartbeat = sctp.param_heartbeat(b'\x01\x02\x03\x04')

        self.heartbeat = sctp.chunk_heartbeat(info=self.p_heartbeat)

        self.chunks = [self.heartbeat]

        self.sc = sctp.sctp(
            self.src_port, self.dst_port, self.vtag, self.csum,
            self.chunks)

        self.buf += b'\x04\x00\x00\x0c' + \
            b'\x00\x01\x00\x08' + \
            b'\x01\x02\x03\x04'

    def setUp_with_heartbeat_ack(self):
        self.flags = 0
        self.length = 4 + 12

        self.p_heartbeat = sctp.param_heartbeat(
            b'\xff\xee\xdd\xcc\xbb\xaa\x99\x88')

        self.heartbeat_ack = sctp.chunk_heartbeat_ack(info=self.p_heartbeat)

        self.chunks = [self.heartbeat_ack]

        self.sc = sctp.sctp(
            self.src_port, self.dst_port, self.vtag, self.csum,
            self.chunks)

        self.buf += b'\x05\x00\x00\x10' + \
            b'\x00\x01\x00\x0c' + \
            b'\xff\xee\xdd\xcc\xbb\xaa\x99\x88'

    def setUp_with_abort(self):
        self.tflag = 0
        self.length = 4 + 8 + 16 + 8 + 4 + 20 + 8 + 4 + 8 + 8 + 4 + 12 \
            + 20 + 20

        self.c_invalid_stream_id = sctp.cause_invalid_stream_id(4096)
        self.c_missing_param = sctp.cause_missing_param(
            [sctp.PTYPE_IPV4, sctp.PTYPE_IPV6,
             sctp.PTYPE_COOKIE_PRESERVE, sctp.PTYPE_HOST_ADDR])
        self.c_stale_cookie = sctp.cause_stale_cookie(b'\x00\x00\x13\x88')
        self.c_out_of_resource = sctp.cause_out_of_resource()
        self.c_unresolvable_addr = sctp.cause_unresolvable_addr(
            sctp.param_host_addr(b'test host\x00'))
        self.c_unrecognized_chunk = sctp.cause_unrecognized_chunk(
            b'\xff\x00\x00\x04')
        self.c_invalid_param = sctp.cause_invalid_param()
        self.c_unrecognized_param = sctp.cause_unrecognized_param(
            b'\xff\xff\x00\x04')
        self.c_no_userdata = sctp.cause_no_userdata(b'\x00\x01\xe2\x40')
        self.c_cookie_while_shutdown = sctp.cause_cookie_while_shutdown()
        self.c_restart_with_new_addr = sctp.cause_restart_with_new_addr(
            sctp.param_ipv4('192.168.1.1'))
        self.c_user_initiated_abort = sctp.cause_user_initiated_abort(
            b'Key Interrupt.\x00')
        self.c_protocol_violation = sctp.cause_protocol_violation(
            b'Unknown reason.\x00')

        self.causes = [
            self.c_invalid_stream_id, self.c_missing_param,
            self.c_stale_cookie, self.c_out_of_resource,
            self.c_unresolvable_addr, self.c_unrecognized_chunk,
            self.c_invalid_param, self.c_unrecognized_param,
            self.c_no_userdata, self.c_cookie_while_shutdown,
            self.c_restart_with_new_addr, self.c_user_initiated_abort,
            self.c_protocol_violation]

        self.abort = sctp.chunk_abort(causes=self.causes)

        self.chunks = [self.abort]

        self.sc = sctp.sctp(
            self.src_port, self.dst_port, self.vtag, self.csum,
            self.chunks)

        self.buf += b'\x06\x00\x00\x90' + \
            b'\x00\x01\x00\x08\x10\x00\x00\x00' + \
            b'\x00\x02\x00\x10\x00\x00\x00\x04' + \
            b'\x00\x05\x00\x06\x00\x09\x00\x0b' + \
            b'\x00\x03\x00\x08\x00\x00\x13\x88' + \
            b'\x00\x04\x00\x04' + \
            b'\x00\x05\x00\x14' + \
            b'\x00\x0b\x00\x0e' + \
            b'\x74\x65\x73\x74\x20\x68\x6f\x73\x74\x00\x00\x00' + \
            b'\x00\x06\x00\x08\xff\x00\x00\x04' + \
            b'\x00\x07\x00\x04' + \
            b'\x00\x08\x00\x08\xff\xff\x00\x04' + \
            b'\x00\x09\x00\x08\x00\x01\xe2\x40' + \
            b'\x00\x0a\x00\x04' + \
            b'\x00\x0b\x00\x0c' + \
            b'\x00\x05\x00\x08\xc0\xa8\x01\x01' + \
            b'\x00\x0c\x00\x13' + \
            b'\x4b\x65\x79\x20\x49\x6e\x74\x65' + \
            b'\x72\x72\x75\x70\x74\x2e\x00\x00' + \
            b'\x00\x0d\x00\x14' + \
            b'\x55\x6e\x6b\x6e\x6f\x77\x6e\x20' + \
            b'\x72\x65\x61\x73\x6f\x6e\x2e\x00'

    def setUp_with_shutdown(self):
        self.flags = 0
        self.length = 8
        self.tsn_ack = 123456

        self.shutdown = sctp.chunk_shutdown(tsn_ack=self.tsn_ack)

        self.chunks = [self.shutdown]

        self.sc = sctp.sctp(
            self.src_port, self.dst_port, self.vtag, self.csum,
            self.chunks)

        self.buf += b'\x07\x00\x00\x08\x00\x01\xe2\x40'

    def setUp_with_shutdown_ack(self):
        self.flags = 0
        self.length = 4

        self.shutdown_ack = sctp.chunk_shutdown_ack()

        self.chunks = [self.shutdown_ack]

        self.sc = sctp.sctp(
            self.src_port, self.dst_port, self.vtag, self.csum,
            self.chunks)

        self.buf += b'\x08\x00\x00\x04'

    def setUp_with_error(self):
        self.flags = 0
        self.length = 4 + 8 + 16 + 8 + 4 + 20 + 8 + 4 + 8 + 8 + 4 + 12 \
            + 20 + 20

        self.c_invalid_stream_id = sctp.cause_invalid_stream_id(4096)
        self.c_missing_param = sctp.cause_missing_param(
            [sctp.PTYPE_IPV4, sctp.PTYPE_IPV6,
             sctp.PTYPE_COOKIE_PRESERVE, sctp.PTYPE_HOST_ADDR])
        self.c_stale_cookie = sctp.cause_stale_cookie(b'\x00\x00\x13\x88')
        self.c_out_of_resource = sctp.cause_out_of_resource()
        self.c_unresolvable_addr = sctp.cause_unresolvable_addr(
            sctp.param_host_addr(b'test host\x00'))
        self.c_unrecognized_chunk = sctp.cause_unrecognized_chunk(
            b'\xff\x00\x00\x04')
        self.c_invalid_param = sctp.cause_invalid_param()
        self.c_unrecognized_param = sctp.cause_unrecognized_param(
            b'\xff\xff\x00\x04')
        self.c_no_userdata = sctp.cause_no_userdata(b'\x00\x01\xe2\x40')
        self.c_cookie_while_shutdown = sctp.cause_cookie_while_shutdown()
        self.c_restart_with_new_addr = sctp.cause_restart_with_new_addr(
            sctp.param_ipv4('192.168.1.1'))
        self.c_user_initiated_abort = sctp.cause_user_initiated_abort(
            b'Key Interrupt.\x00')
        self.c_protocol_violation = sctp.cause_protocol_violation(
            b'Unknown reason.\x00')

        self.causes = [
            self.c_invalid_stream_id, self.c_missing_param,
            self.c_stale_cookie, self.c_out_of_resource,
            self.c_unresolvable_addr, self.c_unrecognized_chunk,
            self.c_invalid_param, self.c_unrecognized_param,
            self.c_no_userdata, self.c_cookie_while_shutdown,
            self.c_restart_with_new_addr, self.c_user_initiated_abort,
            self.c_protocol_violation]

        self.error = sctp.chunk_error(causes=self.causes)

        self.chunks = [self.error]

        self.sc = sctp.sctp(
            self.src_port, self.dst_port, self.vtag, self.csum,
            self.chunks)

        self.buf += b'\x09\x00\x00\x90' + \
            b'\x00\x01\x00\x08\x10\x00\x00\x00' + \
            b'\x00\x02\x00\x10\x00\x00\x00\x04' + \
            b'\x00\x05\x00\x06\x00\x09\x00\x0b' + \
            b'\x00\x03\x00\x08\x00\x00\x13\x88' + \
            b'\x00\x04\x00\x04' + \
            b'\x00\x05\x00\x14' + \
            b'\x00\x0b\x00\x0e' + \
            b'\x74\x65\x73\x74\x20\x68\x6f\x73\x74\x00\x00\x00' + \
            b'\x00\x06\x00\x08\xff\x00\x00\x04' + \
            b'\x00\x07\x00\x04' + \
            b'\x00\x08\x00\x08\xff\xff\x00\x04' + \
            b'\x00\x09\x00\x08\x00\x01\xe2\x40' + \
            b'\x00\x0a\x00\x04' + \
            b'\x00\x0b\x00\x0c' + \
            b'\x00\x05\x00\x08\xc0\xa8\x01\x01' + \
            b'\x00\x0c\x00\x13' + \
            b'\x4b\x65\x79\x20\x49\x6e\x74\x65' + \
            b'\x72\x72\x75\x70\x74\x2e\x00\x00' + \
            b'\x00\x0d\x00\x14' + \
            b'\x55\x6e\x6b\x6e\x6f\x77\x6e\x20' + \
            b'\x72\x65\x61\x73\x6f\x6e\x2e\x00'

    def setUp_with_cookie_echo(self):
        self.flags = 0
        self.length = 8
        self.cookie = b'\x12\x34\x56\x78'

        self.cookie_echo = sctp.chunk_cookie_echo(cookie=self.cookie)

        self.chunks = [self.cookie_echo]

        self.sc = sctp.sctp(
            self.src_port, self.dst_port, self.vtag, self.csum,
            self.chunks)

        self.buf += b'\x0a\x00\x00\x08\x12\x34\x56\x78'

    def setUp_with_cookie_ack(self):
        self.flags = 0
        self.length = 4

        self.cookie_ack = sctp.chunk_cookie_ack()

        self.chunks = [self.cookie_ack]

        self.sc = sctp.sctp(
            self.src_port, self.dst_port, self.vtag, self.csum,
            self.chunks)

        self.buf += b'\x0b\x00\x00\x04'

    def setUp_with_ecn_echo(self):
        self.flags = 0
        self.length = 8
        self.low_tsn = 123456

        self.ecn_echo = sctp.chunk_ecn_echo(low_tsn=self.low_tsn)

        self.chunks = [self.ecn_echo]

        self.sc = sctp.sctp(
            self.src_port, self.dst_port, self.vtag, self.csum,
            self.chunks)

        self.buf += b'\x0c\x00\x00\x08\x00\x01\xe2\x40'

    def setUp_with_cwr(self):
        self.flags = 0
        self.length = 8
        self.low_tsn = 123456

        self.cwr = sctp.chunk_cwr(low_tsn=self.low_tsn)

        self.chunks = [self.cwr]

        self.sc = sctp.sctp(
            self.src_port, self.dst_port, self.vtag, self.csum,
            self.chunks)

        self.buf += b'\x0d\x00\x00\x08\x00\x01\xe2\x40'

    def setUp_with_shutdown_complete(self):
        self.tflag = 0
        self.length = 4

        self.shutdown_complete = sctp.chunk_shutdown_complete()

        self.chunks = [self.shutdown_complete]

        self.sc = sctp.sctp(
            self.src_port, self.dst_port, self.vtag, self.csum,
            self.chunks)

        self.buf += b'\x0e\x00\x00\x04'

    def setUp_with_multi_chunks(self):
        self.s_flags = 0
        self.s_length = 16
        self.s_tsn_ack = 123456
        self.s_a_rwnd = 9876
        self.s_gapack_num = 0
        self.s_duptsn_num = 0
        self.s_gapacks = None
        self.s_duptsns = None

        self.sack = sctp.chunk_sack(
            tsn_ack=self.s_tsn_ack, a_rwnd=self.s_a_rwnd)

        self.d1_unordered = 0
        self.d1_begin = 1
        self.d1_end = 0
        self.d1_length = 16 + 10
        self.d1_tsn = 12345
        self.d1_sid = 1
        self.d1_seq = 0
        self.d1_payload_id = 0
        self.d1_payload_data = b'\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a'

        self.data1 = sctp.chunk_data(
            begin=self.d1_begin, tsn=self.d1_tsn, sid=self.d1_sid,
            payload_data=self.d1_payload_data)

        self.d2_unordered = 0
        self.d2_begin = 0
        self.d2_end = 1
        self.d2_length = 16 + 10
        self.d2_tsn = 12346
        self.d2_sid = 1
        self.d2_seq = 1
        self.d2_payload_id = 0
        self.d2_payload_data = b'\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a'

        self.data2 = sctp.chunk_data(
            end=self.d2_end, tsn=self.d2_tsn, sid=self.d2_sid,
            seq=self.d2_seq, payload_data=self.d2_payload_data)

        self.chunks = [self.sack, self.data1, self.data2]

        self.sc = sctp.sctp(
            self.src_port, self.dst_port, self.vtag, self.csum,
            self.chunks)

        self.buf += b'\x03\x00\x00\x10\x00\x01\xe2\x40' + \
            b'\x00\x00\x26\x94\x00\x00\x00\x00' + \
            b'\x00\x02\x00\x1a\x00\x00\x30\x39\x00\x01\x00\x00' + \
            b'\x00\x00\x00\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a' + \
            b'\x00\x01\x00\x1a\x00\x00\x30\x3a\x00\x01\x00\x01' + \
            b'\x00\x00\x00\x00\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a'

    def tearDown(self):
        pass

    def test_init(self):
        self.assertEqual(self.src_port, self.sc.src_port)
        self.assertEqual(self.dst_port, self.sc.dst_port)
        self.assertEqual(self.vtag, self.sc.vtag)
        self.assertEqual(self.csum, self.sc.csum)
        self.assertEqual(self.chunks, self.sc.chunks)

    def test_init_with_data(self):
        self.setUp_with_data()
        self.test_init()

    def test_init_with_init(self):
        self.setUp_with_init()
        self.test_init()

    def test_init_with_init_ack(self):
        self.setUp_with_init_ack()
        self.test_init()

    def test_init_with_sack(self):
        self.setUp_with_sack()
        self.test_init()

    def test_init_with_heartbeat(self):
        self.setUp_with_heartbeat()
        self.test_init()

    def test_init_with_heartbeat_ack(self):
        self.setUp_with_heartbeat_ack()
        self.test_init()

    def test_init_with_abort(self):
        self.setUp_with_abort()
        self.test_init()

    def test_init_with_shutdown(self):
        self.setUp_with_shutdown()
        self.test_init()

    def test_init_with_shutdown_ack(self):
        self.setUp_with_shutdown_ack()
        self.test_init()

    def test_init_with_error(self):
        self.setUp_with_error()
        self.test_init()

    def test_init_with_cookie_echo(self):
        self.setUp_with_cookie_echo()
        self.test_init()

    def test_init_with_cookie_ack(self):
        self.setUp_with_cookie_ack()
        self.test_init()

    def test_init_with_ecn_echo(self):
        self.setUp_with_ecn_echo()
        self.test_init()

    def test_init_with_cwr(self):
        self.setUp_with_cwr()
        self.test_init()

    def test_init_with_shutdown_complete(self):
        self.setUp_with_shutdown_complete()
        self.test_init()

    def test_init_with_multi_chunks(self):
        self.setUp_with_multi_chunks()
        self.test_init()

    def test_parser(self):
        _res = self.sc.parser(bytes(self.buf))
        if type(_res) is tuple:
            res = _res[0]
        else:
            res = _res
        # to calculate the lengths of parameters.
        self.sc.serialize(None, None)

        self.assertEqual(self.src_port, res.src_port)
        self.assertEqual(self.dst_port, res.dst_port)
        self.assertEqual(self.vtag, res.vtag)
        self.assertEqual(self.csum, res.csum)
        self.assertEqual(str(self.chunks), str(res.chunks))

    def test_parser_with_data(self):
        self.setUp_with_data()
        self.test_parser()

    def test_parser_with_init(self):
        self.setUp_with_init()
        self.test_parser()

    def test_parser_with_init_ack(self):
        self.setUp_with_init_ack()
        self.test_parser()

    def test_parser_with_sack(self):
        self.setUp_with_sack()
        self.test_parser()

    def test_parser_with_heartbeat(self):
        self.setUp_with_heartbeat()
        self.test_parser()

    def test_parser_with_heartbeat_ack(self):
        self.setUp_with_heartbeat_ack()
        self.test_parser()

    def test_parser_with_abort(self):
        self.setUp_with_abort()
        self.test_parser()

    def test_parser_with_shutdown(self):
        self.setUp_with_shutdown()
        self.test_parser()

    def test_parser_with_shutdown_ack(self):
        self.setUp_with_shutdown_ack()
        self.test_parser()

    def test_parser_with_error(self):
        self.setUp_with_error()
        self.test_parser()

    def test_parser_with_cookie_echo(self):
        self.setUp_with_cookie_echo()
        self.test_parser()

    def test_parser_with_cookie_ack(self):
        self.setUp_with_cookie_ack()
        self.test_parser()

    def test_parser_with_ecn_echo(self):
        self.setUp_with_ecn_echo()
        self.test_parser()

    def test_parser_with_cwr(self):
        self.setUp_with_cwr()
        self.test_parser()

    def test_parser_with_shutdown_complete(self):
        self.setUp_with_shutdown_complete()
        self.test_parser()

    def test_parser_with_multi_chunks(self):
        self.setUp_with_multi_chunks()
        self.test_parser()

    def _test_serialize(self):
        buf = self.sc.serialize(bytearray(), None)
        res = struct.unpack_from(sctp.sctp._PACK_STR, buf)
        self.assertEqual(self.src_port, res[0])
        self.assertEqual(self.dst_port, res[1])
        self.assertEqual(self.vtag, res[2])
        # skip compare checksum
        # self.assertEqual(self.csum, res[3])

        return buf[sctp.sctp._MIN_LEN:]

    def test_serialize(self):
        self._test_serialize()

    def test_serialize_with_data(self):
        self.setUp_with_data()
        buf = self._test_serialize()
        res = struct.unpack_from(sctp.chunk_data._PACK_STR, buf)
        self.assertEqual(sctp.chunk_data.chunk_type(), res[0])
        flags = (
            (self.unordered << 2) |
            (self.begin << 1) |
            (self.end << 0))
        self.assertEqual(flags, res[1])
        self.assertEqual(self.length, res[2])
        self.assertEqual(self.tsn, res[3])
        self.assertEqual(self.sid, res[4])
        self.assertEqual(self.seq, res[5])
        self.assertEqual(self.payload_id, res[6])
        self.assertEqual(self.payload_data, buf[sctp.chunk_data._MIN_LEN:])

    def test_serialize_with_init(self):
        self.setUp_with_init()
        buf = self._test_serialize()
        res = struct.unpack_from(sctp.chunk_init._PACK_STR, buf)
        self.assertEqual(sctp.chunk_init.chunk_type(), res[0])
        self.assertEqual(self.flags, res[1])
        self.assertEqual(self.length, res[2])
        self.assertEqual(self.init_tag, res[3])
        self.assertEqual(self.a_rwnd, res[4])
        self.assertEqual(self.os, res[5])
        self.assertEqual(self.mis, res[6])
        self.assertEqual(self.i_tsn, res[7])

        buf = buf[sctp.chunk_init._MIN_LEN:]
        res1 = struct.unpack_from(sctp.param_ipv4._PACK_STR, buf)
        self.assertEqual(sctp.param_ipv4.param_type(), res1[0])
        self.assertEqual(8, res1[1])
        self.assertEqual('192.168.1.1', addrconv.ipv4.bin_to_text(
            buf[sctp.param_ipv4._MIN_LEN:sctp.param_ipv4._MIN_LEN + 4]))

        buf = buf[8:]
        res2 = struct.unpack_from(sctp.param_ipv6._PACK_STR, buf)
        self.assertEqual(sctp.param_ipv6.param_type(), res2[0])
        self.assertEqual(20, res2[1])
        self.assertEqual('fe80::647e:1aff:fec4:8284', addrconv.ipv6.bin_to_text(
            buf[sctp.param_ipv6._MIN_LEN:sctp.param_ipv6._MIN_LEN + 16]))

        buf = buf[20:]
        res3 = struct.unpack_from(sctp.param_cookie_preserve._PACK_STR,
                                  buf)
        self.assertEqual(sctp.param_cookie_preserve.param_type(), res3[0])
        self.assertEqual(8, res3[1])
        self.assertEqual(5000, res3[2])

        buf = buf[8:]
        res4 = struct.unpack_from(sctp.param_ecn._PACK_STR, buf)
        self.assertEqual(sctp.param_ecn.param_type(), res4[0])
        self.assertEqual(4, res4[1])

        buf = buf[4:]
        res5 = struct.unpack_from(sctp.param_host_addr._PACK_STR, buf)
        self.assertEqual(sctp.param_host_addr.param_type(), res5[0])
        self.assertEqual(14, res5[1])
        self.assertEqual(b'test host\x00',
            buf[sctp.param_host_addr._MIN_LEN:
                sctp.param_host_addr._MIN_LEN + 10])

        buf = buf[16:]
        res6 = struct.unpack_from(sctp.param_supported_addr._PACK_STR, buf)
        res6 = list(res6)
        self.assertEqual(sctp.param_supported_addr.param_type(), res6[0])
        self.assertEqual(14, res6[1])
        buf = buf[sctp.param_supported_addr._MIN_LEN:]
        offset = 0
        tmplist = []
        while offset < len(buf):
            (tmp, ) = struct.unpack_from('!H', buf, offset)
            tmplist.append(tmp)
            offset += struct.calcsize('!H')
        res6.extend(tmplist)
        self.assertEqual(sctp.PTYPE_IPV4, res6[2])
        self.assertEqual(sctp.PTYPE_IPV6, res6[3])
        self.assertEqual(sctp.PTYPE_COOKIE_PRESERVE, res6[4])
        self.assertEqual(sctp.PTYPE_ECN, res6[5])
        self.assertEqual(sctp.PTYPE_HOST_ADDR, res6[6])

    def test_serialize_with_init_ack(self):
        self.setUp_with_init_ack()
        buf = self._test_serialize()
        res = struct.unpack_from(sctp.chunk_init_ack._PACK_STR, buf)
        self.assertEqual(sctp.chunk_init_ack.chunk_type(), res[0])
        self.assertEqual(self.flags, res[1])
        self.assertEqual(self.length, res[2])
        self.assertEqual(self.init_tag, res[3])
        self.assertEqual(self.a_rwnd, res[4])
        self.assertEqual(self.os, res[5])
        self.assertEqual(self.mis, res[6])
        self.assertEqual(self.i_tsn, res[7])

        buf = buf[sctp.chunk_init_ack._MIN_LEN:]
        res1 = struct.unpack_from(sctp.param_state_cookie._PACK_STR, buf)
        self.assertEqual(sctp.param_state_cookie.param_type(), res1[0])
        self.assertEqual(7, res1[1])
        self.assertEqual(b'\x01\x02\x03',
            buf[sctp.param_state_cookie._MIN_LEN:
                sctp.param_state_cookie._MIN_LEN + 3])

        buf = buf[8:]
        res2 = struct.unpack_from(sctp.param_ipv4._PACK_STR, buf)
        self.assertEqual(sctp.param_ipv4.param_type(), res2[0])
        self.assertEqual(8, res2[1])
        self.assertEqual('192.168.1.1', addrconv.ipv4.bin_to_text(
            buf[sctp.param_ipv4._MIN_LEN:sctp.param_ipv4._MIN_LEN + 4]))

        buf = buf[8:]
        res3 = struct.unpack_from(sctp.param_ipv6._PACK_STR, buf)
        self.assertEqual(sctp.param_ipv6.param_type(), res3[0])
        self.assertEqual(20, res3[1])
        self.assertEqual('fe80::647e:1aff:fec4:8284', addrconv.ipv6.bin_to_text(
            buf[sctp.param_ipv6._MIN_LEN:sctp.param_ipv6._MIN_LEN + 16]))

        buf = buf[20:]
        res4 = struct.unpack_from(
            sctp.param_unrecognized_param._PACK_STR, buf)
        self.assertEqual(sctp.param_unrecognized_param.param_type(), res4[0])
        self.assertEqual(8, res4[1])
        self.assertEqual(b'\xff\xff\x00\x04',
            buf[sctp.param_unrecognized_param._MIN_LEN:
                sctp.param_unrecognized_param._MIN_LEN + 4])

        buf = buf[8:]
        res5 = struct.unpack_from(sctp.param_ecn._PACK_STR, buf)
        self.assertEqual(sctp.param_ecn.param_type(), res5[0])
        self.assertEqual(4, res5[1])

        buf = buf[4:]
        res6 = struct.unpack_from(sctp.param_host_addr._PACK_STR, buf)
        self.assertEqual(sctp.param_host_addr.param_type(), res6[0])
        self.assertEqual(14, res6[1])
        self.assertEqual(b'test host\x00',
            buf[sctp.param_host_addr._MIN_LEN:
                sctp.param_host_addr._MIN_LEN + 10])

    def test_serialize_with_sack(self):
        self.setUp_with_sack()
        buf = self._test_serialize()
        res = struct.unpack_from(sctp.chunk_sack._PACK_STR, buf)
        self.assertEqual(sctp.chunk_sack.chunk_type(), res[0])
        self.assertEqual(self.flags, res[1])
        self.assertEqual(self.length, res[2])
        self.assertEqual(self.tsn_ack, res[3])
        self.assertEqual(self.a_rwnd, res[4])
        self.assertEqual(self.gapack_num, res[5])
        self.assertEqual(self.duptsn_num, res[6])

        buf = buf[sctp.chunk_sack._MIN_LEN:]
        gapacks = []
        for _ in range(self.gapack_num):
            (gap_s, gap_e) = struct.unpack_from(
                sctp.chunk_sack._GAPACK_STR, buf)
            one = [gap_s, gap_e]
            gapacks.append(one)
            buf = buf[sctp.chunk_sack._GAPACK_LEN:]
        duptsns = []
        for _ in range(self.duptsn_num):
            (duptsn, ) = struct.unpack_from(
                sctp.chunk_sack._DUPTSN_STR, buf)
            duptsns.append(duptsn)
            buf = buf[sctp.chunk_sack._DUPTSN_LEN:]
        self.assertEqual(self.gapacks, gapacks)
        self.assertEqual(self.duptsns, duptsns)

    def test_serialize_with_heartbeat(self):
        self.setUp_with_heartbeat()
        buf = self._test_serialize()
        res = struct.unpack_from(sctp.chunk_heartbeat._PACK_STR, buf)
        self.assertEqual(sctp.chunk_heartbeat.chunk_type(), res[0])
        self.assertEqual(self.flags, res[1])
        self.assertEqual(self.length, res[2])

        buf = buf[sctp.chunk_heartbeat._MIN_LEN:]
        res1 = struct.unpack_from(sctp.param_heartbeat._PACK_STR, buf)
        self.assertEqual(sctp.param_heartbeat.param_type(), res1[0])
        self.assertEqual(8, res1[1])
        self.assertEqual(b'\x01\x02\x03\x04',
            buf[sctp.param_heartbeat._MIN_LEN:
                sctp.param_heartbeat._MIN_LEN + 4])

    def test_serialize_with_heartbeat_ack(self):
        self.setUp_with_heartbeat_ack()
        buf = self._test_serialize()
        res = struct.unpack_from(sctp.chunk_heartbeat_ack._PACK_STR, buf)
        self.assertEqual(sctp.chunk_heartbeat_ack.chunk_type(), res[0])
        self.assertEqual(self.flags, res[1])
        self.assertEqual(self.length, res[2])

        buf = buf[sctp.chunk_heartbeat_ack._MIN_LEN:]
        res1 = struct.unpack_from(sctp.param_heartbeat._PACK_STR, buf)
        self.assertEqual(sctp.param_heartbeat.param_type(), res1[0])
        self.assertEqual(12, res1[1])
        self.assertEqual(b'\xff\xee\xdd\xcc\xbb\xaa\x99\x88',
            buf[sctp.param_heartbeat._MIN_LEN:
                sctp.param_heartbeat._MIN_LEN + 8])

    def test_serialize_with_abort(self):
        self.setUp_with_abort()
        buf = self._test_serialize()
        res = struct.unpack_from(sctp.chunk_abort._PACK_STR, buf)
        self.assertEqual(sctp.chunk_abort.chunk_type(), res[0])
        flags = self.tflag << 0
        self.assertEqual(flags, res[1])
        self.assertEqual(self.length, res[2])

        buf = buf[sctp.chunk_abort._MIN_LEN:]
        res1 = struct.unpack_from(sctp.cause_invalid_stream_id._PACK_STR, buf)
        self.assertEqual(sctp.cause_invalid_stream_id.cause_code(), res1[0])
        self.assertEqual(8, res1[1])
        self.assertEqual(4096, res1[2])

        buf = buf[8:]
        res2 = struct.unpack_from(sctp.cause_missing_param._PACK_STR, buf)
        self.assertEqual(sctp.cause_missing_param.cause_code(), res2[0])
        self.assertEqual(16, res2[1])
        self.assertEqual(4, res2[2])
        types = []
        for count in range(4):
            (tmp, ) = struct.unpack_from(
                '!H', buf, sctp.cause_missing_param._MIN_LEN + 2 * count)
            types.append(tmp)
        self.assertEqual(str([sctp.PTYPE_IPV4, sctp.PTYPE_IPV6,
                 sctp.PTYPE_COOKIE_PRESERVE, sctp.PTYPE_HOST_ADDR]),
            str(types))

        buf = buf[16:]
        res3 = struct.unpack_from(sctp.cause_stale_cookie._PACK_STR, buf)
        self.assertEqual(sctp.cause_stale_cookie.cause_code(), res3[0])
        self.assertEqual(8, res3[1])
        self.assertEqual(b'\x00\x00\x13\x88',
            buf[sctp.cause_stale_cookie._MIN_LEN:
                sctp.cause_stale_cookie._MIN_LEN + 4])

        buf = buf[8:]
        res4 = struct.unpack_from(sctp.cause_out_of_resource._PACK_STR, buf)
        self.assertEqual(sctp.cause_out_of_resource.cause_code(), res4[0])
        self.assertEqual(4, res4[1])

        buf = buf[4:]
        res5 = struct.unpack_from(
            sctp.cause_unresolvable_addr._PACK_STR, buf)
        self.assertEqual(sctp.cause_unresolvable_addr.cause_code(), res5[0])
        self.assertEqual(20, res5[1])
        self.assertEqual(b'\x00\x0b\x00\x0e\x74\x65\x73\x74' +
            b'\x20\x68\x6f\x73\x74\x00\x00\x00',
            buf[sctp.cause_unresolvable_addr._MIN_LEN:
                sctp.cause_unresolvable_addr._MIN_LEN + 16])

        buf = buf[20:]
        res6 = struct.unpack_from(
            sctp.cause_unrecognized_chunk._PACK_STR, buf)
        self.assertEqual(sctp.cause_unrecognized_chunk.cause_code(), res6[0])
        self.assertEqual(8, res6[1])
        self.assertEqual(b'\xff\x00\x00\x04',
            buf[sctp.cause_unrecognized_chunk._MIN_LEN:
                sctp.cause_unrecognized_chunk._MIN_LEN + 4])

        buf = buf[8:]
        res7 = struct.unpack_from(sctp.cause_invalid_param._PACK_STR, buf)
        self.assertEqual(sctp.cause_invalid_param.cause_code(), res7[0])
        self.assertEqual(4, res7[1])

        buf = buf[4:]
        res8 = struct.unpack_from(
            sctp.cause_unrecognized_param._PACK_STR, buf)
        self.assertEqual(sctp.cause_unrecognized_param.cause_code(), res8[0])
        self.assertEqual(8, res8[1])
        self.assertEqual(b'\xff\xff\x00\x04',
            buf[sctp.cause_unrecognized_param._MIN_LEN:
                sctp.cause_unrecognized_param._MIN_LEN + 4])

        buf = buf[8:]
        res9 = struct.unpack_from(sctp.cause_no_userdata._PACK_STR, buf)
        self.assertEqual(sctp.cause_no_userdata.cause_code(), res9[0])
        self.assertEqual(8, res9[1])
        self.assertEqual(b'\x00\x01\xe2\x40',
            buf[sctp.cause_no_userdata._MIN_LEN:
                sctp.cause_no_userdata._MIN_LEN + 4])

        buf = buf[8:]
        res10 = struct.unpack_from(
            sctp.cause_cookie_while_shutdown._PACK_STR, buf)
        self.assertEqual(sctp.cause_cookie_while_shutdown.cause_code(), res10[0])
        self.assertEqual(4, res10[1])

        buf = buf[4:]
        res11 = struct.unpack_from(
            sctp.cause_restart_with_new_addr._PACK_STR, buf)
        self.assertEqual(sctp.cause_restart_with_new_addr.cause_code(), res11[0])
        self.assertEqual(12, res11[1])
        self.assertEqual(b'\x00\x05\x00\x08\xc0\xa8\x01\x01',
            buf[sctp.cause_restart_with_new_addr._MIN_LEN:
                sctp.cause_restart_with_new_addr._MIN_LEN + 8])

        buf = buf[12:]
        res12 = struct.unpack_from(
            sctp.cause_user_initiated_abort._PACK_STR, buf)
        self.assertEqual(sctp.cause_user_initiated_abort.cause_code(), res12[0])
        self.assertEqual(19, res12[1])
        self.assertEqual(b'Key Interrupt.\x00',
            buf[sctp.cause_user_initiated_abort._MIN_LEN:
                sctp.cause_user_initiated_abort._MIN_LEN + 15])

        buf = buf[20:]
        res13 = struct.unpack_from(
            sctp.cause_protocol_violation._PACK_STR, buf)
        self.assertEqual(sctp.cause_protocol_violation.cause_code(), res13[0])
        self.assertEqual(20, res13[1])
        self.assertEqual(b'Unknown reason.\x00',
            buf[sctp.cause_protocol_violation._MIN_LEN:
                sctp.cause_protocol_violation._MIN_LEN + 16])

    def test_serialize_with_shutdown(self):
        self.setUp_with_shutdown()
        buf = self._test_serialize()
        res = struct.unpack_from(sctp.chunk_shutdown._PACK_STR, buf)
        self.assertEqual(sctp.chunk_shutdown.chunk_type(), res[0])
        self.assertEqual(self.flags, res[1])
        self.assertEqual(self.length, res[2])
        self.assertEqual(self.tsn_ack, res[3])

    def test_serialize_with_shutdown_ack(self):
        self.setUp_with_shutdown_ack()
        buf = self._test_serialize()
        res = struct.unpack_from(sctp.chunk_shutdown_ack._PACK_STR, buf)
        self.assertEqual(sctp.chunk_shutdown_ack.chunk_type(), res[0])
        self.assertEqual(self.flags, res[1])
        self.assertEqual(self.length, res[2])

    def test_serialize_with_error(self):
        self.setUp_with_error()
        buf = self._test_serialize()
        res = struct.unpack_from(sctp.chunk_error._PACK_STR, buf)
        self.assertEqual(sctp.chunk_error.chunk_type(), res[0])
        self.assertEqual(self.flags, res[1])
        self.assertEqual(self.length, res[2])

        buf = buf[sctp.chunk_error._MIN_LEN:]
        res1 = struct.unpack_from(sctp.cause_invalid_stream_id._PACK_STR, buf)
        self.assertEqual(sctp.cause_invalid_stream_id.cause_code(), res1[0])
        self.assertEqual(8, res1[1])
        self.assertEqual(4096, res1[2])

        buf = buf[8:]
        res2 = struct.unpack_from(sctp.cause_missing_param._PACK_STR, buf)
        self.assertEqual(sctp.cause_missing_param.cause_code(), res2[0])
        self.assertEqual(16, res2[1])
        self.assertEqual(4, res2[2])
        types = []
        for count in range(4):
            (tmp, ) = struct.unpack_from(
                '!H', buf, sctp.cause_missing_param._MIN_LEN + 2 * count)
            types.append(tmp)
        self.assertEqual(str([sctp.PTYPE_IPV4, sctp.PTYPE_IPV6,
                 sctp.PTYPE_COOKIE_PRESERVE, sctp.PTYPE_HOST_ADDR]),
            str(types))

        buf = buf[16:]
        res3 = struct.unpack_from(sctp.cause_stale_cookie._PACK_STR, buf)
        self.assertEqual(sctp.cause_stale_cookie.cause_code(), res3[0])
        self.assertEqual(8, res3[1])
        self.assertEqual(b'\x00\x00\x13\x88',
            buf[sctp.cause_stale_cookie._MIN_LEN:
                sctp.cause_stale_cookie._MIN_LEN + 4])

        buf = buf[8:]
        res4 = struct.unpack_from(sctp.cause_out_of_resource._PACK_STR, buf)
        self.assertEqual(sctp.cause_out_of_resource.cause_code(), res4[0])
        self.assertEqual(4, res4[1])

        buf = buf[4:]
        res5 = struct.unpack_from(
            sctp.cause_unresolvable_addr._PACK_STR, buf)
        self.assertEqual(sctp.cause_unresolvable_addr.cause_code(), res5[0])
        self.assertEqual(20, res5[1])
        self.assertEqual(b'\x00\x0b\x00\x0e\x74\x65\x73\x74' +
            b'\x20\x68\x6f\x73\x74\x00\x00\x00',
            buf[sctp.cause_unresolvable_addr._MIN_LEN:
                sctp.cause_unresolvable_addr._MIN_LEN + 16])

        buf = buf[20:]
        res6 = struct.unpack_from(
            sctp.cause_unrecognized_chunk._PACK_STR, buf)
        self.assertEqual(sctp.cause_unrecognized_chunk.cause_code(), res6[0])
        self.assertEqual(8, res6[1])
        self.assertEqual(b'\xff\x00\x00\x04',
            buf[sctp.cause_unrecognized_chunk._MIN_LEN:
                sctp.cause_unrecognized_chunk._MIN_LEN + 4])

        buf = buf[8:]
        res7 = struct.unpack_from(sctp.cause_invalid_param._PACK_STR, buf)
        self.assertEqual(sctp.cause_invalid_param.cause_code(), res7[0])
        self.assertEqual(4, res7[1])

        buf = buf[4:]
        res8 = struct.unpack_from(
            sctp.cause_unrecognized_param._PACK_STR, buf)
        self.assertEqual(sctp.cause_unrecognized_param.cause_code(), res8[0])
        self.assertEqual(8, res8[1])
        self.assertEqual(b'\xff\xff\x00\x04',
            buf[sctp.cause_unrecognized_param._MIN_LEN:
                sctp.cause_unrecognized_param._MIN_LEN + 4])

        buf = buf[8:]
        res9 = struct.unpack_from(sctp.cause_no_userdata._PACK_STR, buf)
        self.assertEqual(sctp.cause_no_userdata.cause_code(), res9[0])
        self.assertEqual(8, res9[1])
        self.assertEqual(b'\x00\x01\xe2\x40',
            buf[sctp.cause_no_userdata._MIN_LEN:
                sctp.cause_no_userdata._MIN_LEN + 4])

        buf = buf[8:]
        res10 = struct.unpack_from(
            sctp.cause_cookie_while_shutdown._PACK_STR, buf)
        self.assertEqual(sctp.cause_cookie_while_shutdown.cause_code(), res10[0])
        self.assertEqual(4, res10[1])

        buf = buf[4:]
        res11 = struct.unpack_from(
            sctp.cause_restart_with_new_addr._PACK_STR, buf)
        self.assertEqual(sctp.cause_restart_with_new_addr.cause_code(), res11[0])
        self.assertEqual(12, res11[1])
        self.assertEqual(b'\x00\x05\x00\x08\xc0\xa8\x01\x01',
            buf[sctp.cause_restart_with_new_addr._MIN_LEN:
                sctp.cause_restart_with_new_addr._MIN_LEN + 8])

        buf = buf[12:]
        res12 = struct.unpack_from(
            sctp.cause_user_initiated_abort._PACK_STR, buf)
        self.assertEqual(sctp.cause_user_initiated_abort.cause_code(), res12[0])
        self.assertEqual(19, res12[1])
        self.assertEqual(b'Key Interrupt.\x00',
            buf[sctp.cause_user_initiated_abort._MIN_LEN:
                sctp.cause_user_initiated_abort._MIN_LEN + 15])

        buf = buf[20:]
        res13 = struct.unpack_from(
            sctp.cause_protocol_violation._PACK_STR, buf)
        self.assertEqual(sctp.cause_protocol_violation.cause_code(), res13[0])
        self.assertEqual(20, res13[1])
        self.assertEqual(b'Unknown reason.\x00',
            buf[sctp.cause_protocol_violation._MIN_LEN:
                sctp.cause_protocol_violation._MIN_LEN + 16])

    def test_serialize_with_cookie_echo(self):
        self.setUp_with_cookie_echo()
        buf = self._test_serialize()
        res = struct.unpack_from(sctp.chunk_cookie_echo._PACK_STR, buf)
        self.assertEqual(sctp.chunk_cookie_echo.chunk_type(), res[0])
        self.assertEqual(self.flags, res[1])
        self.assertEqual(self.length, res[2])
        self.assertEqual(self.cookie,
            buf[sctp.chunk_cookie_echo._MIN_LEN:
                sctp.chunk_cookie_echo._MIN_LEN + 4])

    def test_serialize_with_cookie_ack(self):
        self.setUp_with_cookie_ack()
        buf = self._test_serialize()
        res = struct.unpack_from(sctp.chunk_cookie_ack._PACK_STR, buf)
        self.assertEqual(sctp.chunk_cookie_ack.chunk_type(), res[0])
        self.assertEqual(self.flags, res[1])
        self.assertEqual(self.length, res[2])

    def test_serialize_with_ecn_echo(self):
        self.setUp_with_ecn_echo()
        buf = self._test_serialize()
        res = struct.unpack_from(sctp.chunk_ecn_echo._PACK_STR, buf)
        self.assertEqual(sctp.chunk_ecn_echo.chunk_type(), res[0])
        self.assertEqual(self.flags, res[1])
        self.assertEqual(self.length, res[2])
        self.assertEqual(self.low_tsn, res[3])

    def test_serialize_with_cwr(self):
        self.setUp_with_cwr()
        buf = self._test_serialize()
        res = struct.unpack_from(sctp.chunk_cwr._PACK_STR, buf)
        self.assertEqual(sctp.chunk_cwr.chunk_type(), res[0])
        self.assertEqual(self.flags, res[1])
        self.assertEqual(self.length, res[2])
        self.assertEqual(self.low_tsn, res[3])

    def test_serialize_with_shutdown_complete(self):
        self.setUp_with_shutdown_complete()
        buf = self._test_serialize()
        res = struct.unpack_from(
            sctp.chunk_shutdown_complete._PACK_STR, buf)
        self.assertEqual(sctp.chunk_shutdown_complete.chunk_type(), res[0])
        flags = self.tflag << 0
        self.assertEqual(flags, res[1])
        self.assertEqual(self.length, res[2])

    def test_serialize_with_multi_chunks(self):
        self.setUp_with_multi_chunks()
        buf = self._test_serialize()
        res = struct.unpack_from(sctp.chunk_sack._PACK_STR, buf)
        self.assertEqual(sctp.chunk_sack.chunk_type(), res[0])
        self.assertEqual(self.s_flags, res[1])
        self.assertEqual(self.s_length, res[2])
        self.assertEqual(self.s_tsn_ack, res[3])
        self.assertEqual(self.s_a_rwnd, res[4])
        self.assertEqual(self.s_gapack_num, res[5])
        self.assertEqual(self.s_duptsn_num, res[6])

        buf = buf[self.s_length:]
        res = struct.unpack_from(sctp.chunk_data._PACK_STR, buf)
        self.assertEqual(sctp.chunk_data.chunk_type(), res[0])
        d1_flags = (
            (self.d1_unordered << 2) |
            (self.d1_begin << 1) |
            (self.d1_end << 0))
        self.assertEqual(d1_flags, res[1])
        self.assertEqual(self.d1_length, res[2])
        self.assertEqual(self.d1_tsn, res[3])
        self.assertEqual(self.d1_sid, res[4])
        self.assertEqual(self.d1_seq, res[5])
        self.assertEqual(self.d1_payload_id, res[6])
        self.assertEqual(self.d1_payload_data,
            buf[sctp.chunk_data._MIN_LEN:
                sctp.chunk_data._MIN_LEN + 10])

        buf = buf[self.d1_length:]
        res = struct.unpack_from(sctp.chunk_data._PACK_STR, buf)
        self.assertEqual(sctp.chunk_data.chunk_type(), res[0])
        d2_flags = (
            (self.d2_unordered << 2) |
            (self.d2_begin << 1) |
            (self.d2_end << 0))
        self.assertEqual(d2_flags, res[1])
        self.assertEqual(self.d2_length, res[2])
        self.assertEqual(self.d2_tsn, res[3])
        self.assertEqual(self.d2_sid, res[4])
        self.assertEqual(self.d2_seq, res[5])
        self.assertEqual(self.d2_payload_id, res[6])
        self.assertEqual(self.d2_payload_data,
            buf[sctp.chunk_data._MIN_LEN:
                sctp.chunk_data._MIN_LEN + 10])

    def test_build_sctp(self):
        eth = ethernet.ethernet('00:aa:aa:aa:aa:aa', '00:bb:bb:bb:bb:bb',
                                ether.ETH_TYPE_IP)
        ip4 = ipv4.ipv4(4, 5, 16, 0, 0, 2, 0, 64, inet.IPPROTO_SCTP, 0,
                        '192.168.1.1', '10.144.1.1')
        pkt = eth / ip4 / self.sc

        eth = pkt.get_protocol(ethernet.ethernet)
        self.assertTrue(eth)
        self.assertEqual(eth.ethertype, ether.ETH_TYPE_IP)

        ip4 = pkt.get_protocol(ipv4.ipv4)
        self.assertTrue(ip4)
        self.assertEqual(ip4.proto, inet.IPPROTO_SCTP)

        sc = pkt.get_protocol(sctp.sctp)
        self.assertTrue(sc)
        self.assertEqual(sc, self.sc)

    def test_build_sctp_with_data(self):
        self.setUp_with_data()
        self.test_build_sctp()

    def test_build_sctp_with_init(self):
        self.setUp_with_init()
        self.test_build_sctp()

    def test_build_sctp_with_init_ack(self):
        self.setUp_with_init_ack()
        self.test_build_sctp()

    def test_build_sctp_with_sack(self):
        self.setUp_with_sack()
        self.test_build_sctp()

    def test_build_sctp_with_heartbeat(self):
        self.setUp_with_heartbeat()
        self.test_build_sctp()

    def test_build_sctp_with_heartbeat_ack(self):
        self.setUp_with_heartbeat_ack()
        self.test_build_sctp()

    def test_build_sctp_with_abort(self):
        self.setUp_with_abort()
        self.test_build_sctp()

    def test_build_sctp_with_shutdown(self):
        self.setUp_with_shutdown()
        self.test_build_sctp()

    def test_build_sctp_with_shutdown_ack(self):
        self.setUp_with_shutdown_ack()
        self.test_build_sctp()

    def test_build_sctp_with_error(self):
        self.setUp_with_error()
        self.test_build_sctp()

    def test_build_sctp_with_cookie_echo(self):
        self.setUp_with_cookie_echo()
        self.test_build_sctp()

    def test_build_sctp_with_cookie_ack(self):
        self.setUp_with_cookie_ack()
        self.test_build_sctp()

    def test_build_sctp_with_ecn_echo(self):
        self.setUp_with_ecn_echo()
        self.test_build_sctp()

    def test_build_sctp_with_cwr(self):
        self.setUp_with_cwr()
        self.test_build_sctp()

    def test_build_sctp_with_shutdown_complete(self):
        self.setUp_with_shutdown_complete()
        self.test_build_sctp()

    def tset_build_sctp_with_multi_chunks(self):
        self.setUp_with_multi_chunks()
        self.test_build_sctp()

    def test_to_string(self):
        sctp_values = {'src_port': self.src_port,
                       'dst_port': self.dst_port,
                       'vtag': self.vtag,
                       'csum': self.csum,
                       'chunks': self.chunks}
        _sctp_str = ','.join(['%s=%s' % (k, sctp_values[k])
                              for k, _ in inspect.getmembers(self.sc)
                              if k in sctp_values])
        sctp_str = '%s(%s)' % (sctp.sctp.__name__, _sctp_str)

        self.assertEqual(str(self.sc), sctp_str)
        self.assertEqual(repr(self.sc), sctp_str)

    def test_to_string_with_data(self):
        self.setUp_with_data()
        self.test_to_string()

    def test_to_string_with_init(self):
        self.setUp_with_init()
        self.test_to_string()

    def test_to_string_with_init_ack(self):
        self.setUp_with_init_ack()
        self.test_to_string()

    def test_to_string_with_sack(self):
        self.setUp_with_sack()
        self.test_to_string()

    def test_to_string_with_heartbeat(self):
        self.setUp_with_heartbeat()
        self.test_to_string()

    def test_to_string_with_heartbeat_ack(self):
        self.setUp_with_heartbeat_ack()
        self.test_to_string()

    def test_to_string_with_abort(self):
        self.setUp_with_abort()
        self.test_to_string()

    def test_to_string_with_shutdown(self):
        self.setUp_with_shutdown()
        self.test_to_string()

    def test_to_string_with_shutdown_ack(self):
        self.setUp_with_shutdown_ack()
        self.test_to_string()

    def test_to_string_with_error(self):
        self.setUp_with_error()
        self.test_to_string()

    def test_to_string_with_cookie_echo(self):
        self.setUp_with_cookie_echo()
        self.test_to_string()

    def test_to_string_with_cookie_ack(self):
        self.setUp_with_cookie_ack()
        self.test_to_string()

    def test_to_string_with_ecn_echo(self):
        self.setUp_with_ecn_echo()
        self.test_to_string()

    def test_to_string_with_cwr(self):
        self.setUp_with_cwr()
        self.test_to_string()

    def test_to_string_with_shutdown_complete(self):
        self.setUp_with_shutdown_complete()
        self.test_to_string()

    def test_to_string_with_multi_chunks(self):
        self.setUp_with_multi_chunks()
        self.test_to_string()

    def test_json(self):
        jsondict = self.sc.to_jsondict()
        sc = sctp.sctp.from_jsondict(jsondict['sctp'])
        self.assertEqual(str(self.sc), str(sc))

    def test_json_with_data(self):
        self.setUp_with_data()
        self.test_json()

    def test_json_with_init(self):
        self.setUp_with_init()
        self.test_json()

    def test_json_with_init_ack(self):
        self.setUp_with_init_ack()
        self.test_json()

    def test_json_with_sack(self):
        self.setUp_with_sack()
        self.test_json()

    def test_json_with_heartbeat(self):
        self.setUp_with_heartbeat()
        self.test_json()

    def test_json_with_heartbeat_ack(self):
        self.setUp_with_heartbeat_ack()
        self.test_json()

    def test_json_with_abort(self):
        self.setUp_with_abort()
        self.test_json()

    def test_json_with_shutdown(self):
        self.setUp_with_shutdown()
        self.test_json()

    def test_json_with_shutdown_ack(self):
        self.setUp_with_shutdown_ack()
        self.test_json()

    def test_json_with_error(self):
        self.setUp_with_error()
        self.test_json()

    def test_json_with_cookie_echo(self):
        self.setUp_with_cookie_echo()
        self.test_json()

    def test_json_with_cookie_ack(self):
        self.setUp_with_cookie_ack()
        self.test_json()

    def test_json_with_ecn_echo(self):
        self.setUp_with_ecn_echo()
        self.test_json()

    def test_json_with_cwr(self):
        self.setUp_with_cwr()
        self.test_json()

    def test_json_with_shutdown_complete(self):
        self.setUp_with_shutdown_complete()
        self.test_json()

    def test_json_with_multi_chunks(self):
        self.setUp_with_multi_chunks()
        self.test_json()



#!/usr/bin/env python3

import sys

from mininet.cli import CLI
from mininet.net import Mininet
from mininet.node import RemoteController
from mininet.node import OVSSwitch
from mininet.node import UserSwitch

from oslo_config import cfg
from os_ken import version
from os_ken.ofproto.ofproto_common import OFP_TCP_PORT


if '__main__' == __name__:

    opts = [
        cfg.StrOpt('switch', default='ovs',
                   help='test switch [ovs|cpqd]'),
        cfg.StrOpt('protocols', default='OpenFlow13',
                   help='"protocols" option for ovs-vsctl (e.g. OpenFlow13)')
    ]
    conf = cfg.ConfigOpts()
    conf.register_cli_opts(opts)
    conf(project='os_ken', version='run_mininet.py %s' % version)
    conf(sys.argv[1:])
    switch_type = {'ovs': OVSSwitch, 'cpqd': UserSwitch}
    switch = switch_type.get(conf.switch, None)
    if switch is None:
        raise ValueError('Invalid switch type. [%s]', conf.switch)

    net = Mininet(switch=switch, controller=RemoteController)

    c0 = net.addController('c0', port=OFP_TCP_PORT)

    s1 = net.addSwitch('s1')
    s2 = net.addSwitch('s2')

    net.addLink(s1, s2)
    net.addLink(s1, s2)
    net.addLink(s1, s2)

    net.start()

    if conf.switch == 'ovs':
        s1.cmd('ovs-vsctl set Bridge s1 protocols=%s' % conf.protocols)
        s2.cmd('ovs-vsctl set Bridge s2 protocols=%s' % conf.protocols)

    CLI(net)

    net.stop()



# Copyright (C) 2014 Xinguard, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import logging
import struct
import inspect

from os_ken.ofproto import ether
from os_ken.ofproto import inet
from os_ken.lib.packet import packet
from os_ken.lib.packet import ethernet
from os_ken.lib.packet import ipv4
from os_ken.lib.packet import udp
from os_ken.lib.packet import bfd
from os_ken.lib import addrconv

LOG = logging.getLogger(__name__)


class TestBFD(unittest.TestCase):
    def setUp(self):
        # BFD packet without authentication.
        self.data = b'\xb0\xa8\x6e\x18\xb8\x08\x64\x87' \
                    + b'\x88\xe9\xcb\xc8\x08\x00\x45\xc0' \
                    + b'\x00\x34\x68\x49\x00\x00\xff\x11' \
                    + b'\xf4\x73\xac\x1c\x03\x01\xac\x1c' \
                    + b'\x03\x02\xc0\x00\x0e\xc8\x00\x20' \
                    + b'\xd9\x02\x21\xc0\x03\x18\x00\x00' \
                    + b'\x00\x06\x00\x00\x00\x07\x00\x00' \
                    + b'\xea\x60\x00\x00\xea\x60\x00\x00' \
                    + b'\x00\x00'

        # BFD packet using simple password authentication.
        self.data_auth_simple = b'\x08\x00\x27\xd1\x95\x7c\x08\x00' \
                                + b'\x27\xed\x54\x41\x08\x00\x45\xc0' \
                                + b'\x00\x3d\x0c\x90\x00\x00\xff\x11' \
                                + b'\xbb\x0b\xc0\xa8\x39\x02\xc0\xa8' \
                                + b'\x39\x01\xc0\x00\x0e\xc8\x00\x29' \
                                + b'\x46\x35\x20\x44\x03\x21\x00\x00' \
                                + b'\x00\x01\x00\x00\x00\x00\x00\x0f' \
                                + b'\x42\x40\x00\x0f\x42\x40\x00\x00' \
                                + b'\x00\x00\x01\x09\x02\x73\x65\x63' \
                                + b'\x72\x65\x74'

        # BFD packet using md5 authentication.
        self.data_auth_md5 = b'\x08\x00\x27\xd1\x95\x7c\x08\x00' \
                             + b'\x27\xed\x54\x41\x08\x00\x45\xc0' \
                             + b'\x00\x4c\x0c\x44\x00\x00\xff\x11' \
                             + b'\xbb\x48\xc0\xa8\x39\x02\xc0\xa8' \
                             + b'\x39\x01\xc0\x00\x0e\xc8\x00\x38' \
                             + b'\x51\xbc\x20\x44\x03\x30\x00\x00' \
                             + b'\x00\x01\x00\x00\x00\x00\x00\x0f' \
                             + b'\x42\x40\x00\x0f\x42\x40\x00\x00' \
                             + b'\x00\x00\x02\x18\x02\x00\x00\x00' \
                             + b'\x41\xdb\x66\xa8\xf9\x25\x5a\x8b' \
                             + b'\xcb\x7e\x4b\xec\x25\xa6\x2c\x23' \
                             + b'\xda\x0f'

        # BFD packet using SHA1 authentication.
        self.data_auth_sha1 = b'\x08\x00\x27\xd1\x95\x7c\x08\x00' \
                              + b'\x27\xed\x54\x41\x08\x00\x45\xc0' \
                              + b'\x00\x50\x0b\x90\x00\x00\xff\x11' \
                              + b'\xbb\xf8\xc0\xa8\x39\x02\xc0\xa8' \
                              + b'\x39\x01\xc0\x00\x0e\xc8\x00\x3c' \
                              + b'\xb9\x92\x20\x44\x03\x34\x00\x00' \
                              + b'\x00\x01\x00\x00\x00\x00\x00\x0f' \
                              + b'\x42\x40\x00\x0f\x42\x40\x00\x00' \
                              + b'\x00\x00\x04\x1c\x02\x00\x00\x00' \
                              + b'\x41\xb1\x46\x20\x10\x81\x03\xd7' \
                              + b'\xf4\xde\x87\x61\x4c\x24\x61\x1f' \
                              + b'\x3c\xc1\x6a\x00\x69\x23'

        # BFD Key chain {auth_key_id: auth_key/password}
        self.auth_keys = {2: b"secret"}

    def tearDown(self):
        pass

    def test_parse(self):
        buf = self.data
        pkt = packet.Packet(buf)
        i = iter(pkt)

        self.assertIsInstance(next(i), ethernet.ethernet)
        self.assertIsInstance(next(i), ipv4.ipv4)
        self.assertIsInstance(next(i), udp.udp)
        self.assertIsInstance(bfd.bfd.parser(next(i))[0], bfd.bfd)

    def test_parse_with_auth_simple(self):
        buf = self.data_auth_simple
        pkt = packet.Packet(buf)
        i = iter(pkt)

        self.assertIsInstance(next(i), ethernet.ethernet)
        self.assertIsInstance(next(i), ipv4.ipv4)
        self.assertIsInstance(next(i), udp.udp)

        bfd_obj = bfd.bfd.parser(next(i))[0]
        self.assertIsInstance(bfd_obj, bfd.bfd)
        self.assertIsInstance(bfd_obj.auth_cls, bfd.SimplePassword)
        self.assertTrue(bfd_obj.authenticate(self.auth_keys))

    def test_parse_with_auth_md5(self):
        buf = self.data_auth_md5
        pkt = packet.Packet(buf)
        i = iter(pkt)

        self.assertIsInstance(next(i), ethernet.ethernet)
        self.assertIsInstance(next(i), ipv4.ipv4)
        self.assertIsInstance(next(i), udp.udp)

        bfd_obj = bfd.bfd.parser(next(i))[0]
        self.assertIsInstance(bfd_obj, bfd.bfd)
        self.assertIsInstance(bfd_obj.auth_cls, bfd.KeyedMD5)
        self.assertTrue(bfd_obj.authenticate(self.auth_keys))

    def test_parse_with_auth_sha1(self):
        buf = self.data_auth_sha1
        pkt = packet.Packet(buf)
        i = iter(pkt)

        self.assertIsInstance(next(i), ethernet.ethernet)
        self.assertIsInstance(next(i), ipv4.ipv4)
        self.assertIsInstance(next(i), udp.udp)

        bfd_obj = bfd.bfd.parser(next(i))[0]
        self.assertIsInstance(bfd_obj, bfd.bfd)
        self.assertIsInstance(bfd_obj.auth_cls, bfd.KeyedSHA1)
        self.assertTrue(bfd_obj.authenticate(self.auth_keys))

    def test_serialize(self):
        pkt = packet.Packet()

        eth_pkt = ethernet.ethernet('b0:a8:6e:18:b8:08', '64:87:88:e9:cb:c8')
        pkt.add_protocol(eth_pkt)

        ip_pkt = ipv4.ipv4(src='172.28.3.1', dst='172.28.3.2', tos=192,
                           identification=26697, proto=inet.IPPROTO_UDP)
        pkt.add_protocol(ip_pkt)

        udp_pkt = udp.udp(49152, 3784)
        pkt.add_protocol(udp_pkt)

        bfd_pkt = bfd.bfd(ver=1, diag=bfd.BFD_DIAG_CTRL_DETECT_TIME_EXPIRED,
                          state=bfd.BFD_STATE_UP, detect_mult=3, my_discr=6,
                          your_discr=7, desired_min_tx_interval=60000,
                          required_min_rx_interval=60000,
                          required_min_echo_rx_interval=0)
        pkt.add_protocol(bfd_pkt)

        self.assertEqual(len(pkt.protocols), 4)

        pkt.serialize()
        self.assertEqual(pkt.data, self.data)

    def test_serialize_with_auth_simple(self):
        pkt = packet.Packet()

        eth_pkt = ethernet.ethernet('08:00:27:d1:95:7c', '08:00:27:ed:54:41')
        pkt.add_protocol(eth_pkt)

        ip_pkt = ipv4.ipv4(src='192.168.57.2', dst='192.168.57.1', tos=192,
                           identification=3216, proto=inet.IPPROTO_UDP)
        pkt.add_protocol(ip_pkt)

        udp_pkt = udp.udp(49152, 3784)
        pkt.add_protocol(udp_pkt)

        auth_cls = bfd.SimplePassword(auth_key_id=2,
                                      password=self.auth_keys[2])

        bfd_pkt = bfd.bfd(ver=1, diag=bfd.BFD_DIAG_NO_DIAG,
                          flags=bfd.BFD_FLAG_AUTH_PRESENT,
                          state=bfd.BFD_STATE_DOWN, detect_mult=3, my_discr=1,
                          your_discr=0, desired_min_tx_interval=1000000,
                          required_min_rx_interval=1000000,
                          required_min_echo_rx_interval=0,
                          auth_cls=auth_cls)

        pkt.add_protocol(bfd_pkt)

        self.assertEqual(len(pkt.protocols), 4)

        pkt.serialize()
        self.assertEqual(pkt.data, self.data_auth_simple)

    def test_serialize_with_auth_md5(self):
        pkt = packet.Packet()

        eth_pkt = ethernet.ethernet('08:00:27:d1:95:7c', '08:00:27:ed:54:41')
        pkt.add_protocol(eth_pkt)

        ip_pkt = ipv4.ipv4(src='192.168.57.2', dst='192.168.57.1', tos=192,
                           identification=3140, proto=inet.IPPROTO_UDP)
        pkt.add_protocol(ip_pkt)

        udp_pkt = udp.udp(49152, 3784)
        pkt.add_protocol(udp_pkt)

        auth_cls = bfd.KeyedMD5(auth_key_id=2, seq=16859,
                                auth_key=self.auth_keys[2])

        bfd_pkt = bfd.bfd(ver=1, diag=bfd.BFD_DIAG_NO_DIAG,
                          flags=bfd.BFD_FLAG_AUTH_PRESENT,
                          state=bfd.BFD_STATE_DOWN, detect_mult=3, my_discr=1,
                          your_discr=0, desired_min_tx_interval=1000000,
                          required_min_rx_interval=1000000,
                          required_min_echo_rx_interval=0,
                          auth_cls=auth_cls)

        pkt.add_protocol(bfd_pkt)

        self.assertEqual(len(pkt.protocols), 4)

        pkt.serialize()
        self.assertEqual(pkt.data, self.data_auth_md5)

    def test_serialize_with_auth_sha1(self):
        pkt = packet.Packet()

        eth_pkt = ethernet.ethernet('08:00:27:d1:95:7c', '08:00:27:ed:54:41')
        pkt.add_protocol(eth_pkt)

        ip_pkt = ipv4.ipv4(src='192.168.57.2', dst='192.168.57.1', tos=192,
                           identification=2960, proto=inet.IPPROTO_UDP)
        pkt.add_protocol(ip_pkt)

        udp_pkt = udp.udp(49152, 3784)
        pkt.add_protocol(udp_pkt)

        auth_cls = bfd.KeyedSHA1(auth_key_id=2, seq=16817,
                                 auth_key=self.auth_keys[2])

        bfd_pkt = bfd.bfd(ver=1, diag=bfd.BFD_DIAG_NO_DIAG,
                          flags=bfd.BFD_FLAG_AUTH_PRESENT,
                          state=bfd.BFD_STATE_DOWN, detect_mult=3, my_discr=1,
                          your_discr=0, desired_min_tx_interval=1000000,
                          required_min_rx_interval=1000000,
                          required_min_echo_rx_interval=0,
                          auth_cls=auth_cls)

        pkt.add_protocol(bfd_pkt)

        self.assertEqual(len(pkt.protocols), 4)

        pkt.serialize()
        self.assertEqual(pkt.data, self.data_auth_sha1)

    def test_json(self):
        bfd1 = bfd.bfd(ver=1, diag=bfd.BFD_DIAG_CTRL_DETECT_TIME_EXPIRED,
                       state=bfd.BFD_STATE_UP, detect_mult=3, my_discr=6,
                       your_discr=7, desired_min_tx_interval=60000,
                       required_min_rx_interval=60000,
                       required_min_echo_rx_interval=0)

        jsondict = bfd1.to_jsondict()
        bfd2 = bfd.bfd.from_jsondict(jsondict['bfd'])
        self.assertEqual(str(bfd1), str(bfd2))

    def test_json_with_auth_simple(self):
        auth_cls = bfd.SimplePassword(auth_key_id=2,
                                      password=self.auth_keys[2])

        bfd1 = bfd.bfd(ver=1, diag=bfd.BFD_DIAG_NO_DIAG,
                       flags=bfd.BFD_FLAG_AUTH_PRESENT,
                       state=bfd.BFD_STATE_DOWN, detect_mult=3, my_discr=1,
                       your_discr=0, desired_min_tx_interval=1000000,
                       required_min_rx_interval=1000000,
                       required_min_echo_rx_interval=0,
                       auth_cls=auth_cls)

        jsondict = bfd1.to_jsondict()
        bfd2 = bfd.bfd.from_jsondict(jsondict['bfd'])
        self.assertEqual(str(bfd1), str(bfd2))

    def test_json_with_auth_md5(self):
        auth_cls = bfd.KeyedMD5(auth_key_id=2, seq=16859,
                                auth_key=self.auth_keys[2])

        bfd1 = bfd.bfd(ver=1, diag=bfd.BFD_DIAG_NO_DIAG,
                       flags=bfd.BFD_FLAG_AUTH_PRESENT,
                       state=bfd.BFD_STATE_DOWN, detect_mult=3, my_discr=1,
                       your_discr=0, desired_min_tx_interval=1000000,
                       required_min_rx_interval=1000000,
                       required_min_echo_rx_interval=0,
                       auth_cls=auth_cls)

        jsondict = bfd1.to_jsondict()
        bfd2 = bfd.bfd.from_jsondict(jsondict['bfd'])
        self.assertEqual(str(bfd1), str(bfd2))

    def test_json_with_auth_sha1(self):
        auth_cls = bfd.KeyedSHA1(auth_key_id=2, seq=16859,
                                 auth_key=self.auth_keys[2])

        bfd1 = bfd.bfd(ver=1, diag=bfd.BFD_DIAG_NO_DIAG,
                       flags=bfd.BFD_FLAG_AUTH_PRESENT,
                       state=bfd.BFD_STATE_DOWN, detect_mult=3, my_discr=1,
                       your_discr=0, desired_min_tx_interval=1000000,
                       required_min_rx_interval=1000000,
                       required_min_echo_rx_interval=0,
                       auth_cls=auth_cls)

        jsondict = bfd1.to_jsondict()
        bfd2 = bfd.bfd.from_jsondict(jsondict['bfd'])
        self.assertEqual(str(bfd1), str(bfd2))
